from gettext import gettext as _
from operator import itemgetter
from typing import Optional, Tuple, List

from gi.repository import GObject, GLib, Gio
from blueman.Functions import launch
from blueman.main.PluginManager import PluginManager
from blueman.plugins.AppletPlugin import AppletPlugin


class StatusIconImplementationProvider:
    def on_query_status_icon_implementation(self) -> Tuple[str, int]:
        return "GtkStatusIcon", 0


class StatusIconVisibilityHandler:
    def on_query_force_status_icon_visibility(self) -> bool:
        return False


class StatusIconProvider:
    def on_status_icon_query_icon(self) -> Optional[str]:
        return None


class StatusIcon(AppletPlugin, GObject.GObject):
    __icon__ = "bluetooth-symbolic"
    __depends__ = ["StandardItems", "Menu"]

    visible = None

    visibility_timeout: Optional[int] = None

    _implementations = None

    def on_load(self) -> None:
        GObject.GObject.__init__(self)
        self._tooltip_title = _("Bluetooth Enabled")
        self._tooltip_text = ""

        self.general_config = Gio.Settings(schema_id="org.blueman.general")
        self.general_config.connect("changed::symbolic-status-icons", self.on_symbolic_config_change)

        self.query_visibility(emit=False)

        self.parent.Plugins.connect('plugin-loaded', self._on_plugins_changed)
        self.parent.Plugins.connect('plugin-unloaded', self._on_plugins_changed)

        self._add_dbus_method("GetVisibility", (), "b", lambda: self.visible)
        self._add_dbus_signal("VisibilityChanged", "b")
        self._add_dbus_signal("ToolTipTitleChanged", "s")
        self._add_dbus_signal("ToolTipTextChanged", "s")
        self._add_dbus_method("GetToolTipTitle", (), "s", lambda: self._tooltip_title)
        self._add_dbus_method("GetToolTipText", (), "s", lambda: self._tooltip_text)
        self._add_dbus_signal("IconNameChanged", "s")
        self._add_dbus_method("GetStatusIconImplementations", (), "as", self._get_status_icon_implementations)
        self._add_dbus_method("GetIconName", (), "s", self._get_icon_name)
        self._add_dbus_method("Activate", (), "", self.parent.Plugins.StandardItems.on_devices)

    def query_visibility(self, delay_hiding: bool = False, emit: bool = True) -> None:
        if self.parent.Manager.get_adapters() or \
           any(plugin.on_query_force_status_icon_visibility()
               for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconVisibilityHandler)):
            self.set_visible(True, emit)
        elif not self.visibility_timeout:
            if delay_hiding:
                self.visibility_timeout = GLib.timeout_add(2500, self.on_visibility_timeout)
            else:
                self.set_visible(False, emit)

    def on_visibility_timeout(self) -> bool:
        assert self.visibility_timeout is not None
        GLib.source_remove(self.visibility_timeout)
        self.visibility_timeout = None
        self.query_visibility()
        return False

    def set_visible(self, visible: bool, emit: bool) -> None:
        self.visible = visible
        if emit:
            self._emit_dbus_signal("VisibilityChanged", visible)

    def set_tooltip_title(self, title: str) -> None:
        self._tooltip_title = title
        self._emit_dbus_signal("ToolTipTitleChanged", title)

    def set_tooltip_text(self, text: Optional[str]) -> None:
        self._tooltip_text = "" if text is None else text
        self._emit_dbus_signal("ToolTipTextChanged", self._tooltip_text)

    def on_symbolic_config_change(self, settings: Gio.Settings, key: str) -> None:
        self.icon_should_change()

    def icon_should_change(self) -> None:
        self._emit_dbus_signal("IconNameChanged", self._get_icon_name())
        self.query_visibility()

    def on_adapter_added(self, _path: str) -> None:
        self.query_visibility()

    def on_adapter_removed(self, _path: str) -> None:
        self.query_visibility()

    def on_manager_state_changed(self, state: bool) -> None:
        self.query_visibility()
        if state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _on_plugins_changed(self, _plugins: PluginManager[AppletPlugin], _name: str) -> None:
        implementations = self._get_status_icon_implementations()
        if not self._implementations or self._implementations != implementations:
            self._implementations = implementations

        if self.parent.manager_state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _get_status_icon_implementations(self) -> List[str]:
        return [implementation for implementation, _ in sorted(
            (plugin.on_query_status_icon_implementation()
             for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconImplementationProvider)),
            key=itemgetter(1),
            reverse=True
        )] + ["GtkStatusIcon"]

    def _get_icon_name(self) -> str:
        # default icon name
        name = "blueman-tray"
        for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconProvider):
            icon = plugin.on_status_icon_query_icon()
            if icon is not None:
                # status icon
                name = icon

        # depending on configuration, ensure fullcolor icons..
        name = name.replace("-symbolic", "")
        if self.general_config.get_boolean("symbolic-status-icons"):
            # or symbolic
            name = f"{name}-symbolic"

        return name



from gettext import gettext as _
from random import randint
import logging
import ipaddress
from typing import List, Tuple, cast, Union, TYPE_CHECKING

from blueman.Functions import have, get_local_interfaces
from blueman.main.Builder import Builder
from blueman.plugins.ServicePlugin import ServicePlugin
from blueman.main.DBusProxies import Mechanism
from blueman.main.DBusProxies import AppletService
from blueman.gui.CommonUi import ErrorDialog

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gio, Gtk, GObject

if TYPE_CHECKING:
    from typing_extensions import Literal


class Network(ServicePlugin):
    __plugin_info__ = (_("Network"), "network-workgroup")

    def on_load(self) -> None:

        self._builder = Builder("services-network.ui")
        self.widget = self._builder.get_widget("network", Gtk.Grid)

        self.interfaces: List[Tuple[str, Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface]]] = []
        netifs = get_local_interfaces()
        for iface in netifs:
            if iface != "lo" and iface != "pan1":
                logging.info(iface)
                ipiface = ipaddress.ip_interface('/'.join(cast(Tuple[str, str], netifs[iface])))
                self.interfaces.append((iface, ipiface))

        self.setup_network()
        try:
            self.ip_check()
        except (ValueError, ipaddress.AddressValueError) as e:
            logging.exception(e)

    def on_apply(self) -> None:

        if self.on_query_apply_state():
            logging.info("network apply")

            m = Mechanism()
            nap_enable = self._builder.get_widget("nap-enable", Gtk.CheckButton)
            if nap_enable.props.active:

                if self._builder.get_widget("r_dhcpd", Gtk.RadioButton).props.active:
                    stype = "DhcpdHandler"
                elif self._builder.get_widget("r_dnsmasq", Gtk.RadioButton).props.active:
                    stype = "DnsMasqHandler"
                elif self._builder.get_widget("r_udhcpd", Gtk.RadioButton).props.active:
                    stype = "UdhcpdHandler"

                net_ip = self._builder.get_widget("net_ip", Gtk.Entry)

                try:
                    changed = net_ip.props.text != self.Config["ip4-address"]
                    m.EnableNetwork('(sssb)', net_ip.props.text, "255.255.255.0", stype, changed)

                    if not self.Config["nap-enable"]:
                        self.Config["nap-enable"] = True
                    self.Config["ip4-address"] = net_ip.props.text
                    self.Config["dhcp-handler"] = stype
                except Exception as e:
                    parent = self.widget.get_toplevel()
                    assert isinstance(parent, Gtk.Container)
                    d = ErrorDialog("<b>Failed to apply network settings</b>", excp=e, parent=parent)

                    d.run()
                    d.destroy()
                    return
            else:
                self.Config["nap-enable"] = False
                m.DisableNetwork()

            self.clear_options()

    def ip_check(self) -> None:
        entry = self._builder.get_widget("net_ip", Gtk.Entry)
        try:
            nap_ipiface = ipaddress.ip_interface('/'.join((entry.props.text, '255.255.255.0')))
        except (ValueError, ipaddress.AddressValueError):
            entry.props.secondary_icon_name = "dialog-error"
            entry.props.secondary_icon_tooltip_text = _("Invalid IP address")
            raise

        for iface, ipiface in self.interfaces:
            if nap_ipiface.ip == ipiface.ip:
                error_message = _("IP address conflicts with interface %s which has the same address" % iface)
                tooltip_text = error_message
                entry.props.secondary_icon_name = "dialog-error"
                entry.props.secondary_icon_tooltip_text = tooltip_text
                raise ValueError(error_message)

            elif nap_ipiface.network == ipiface.network:
                tooltip_text = _(
                    "IP address overlaps with subnet of interface %s, which has the following configuration  %s/%s\n"
                    "This may cause incorrect network behavior" % (iface, ipiface.ip, ipiface.netmask))
                entry.props.secondary_icon_name = "dialog-warning"
                entry.props.secondary_icon_tooltip_text = tooltip_text
                return

        entry.props.secondary_icon_name = None

    def on_query_apply_state(self) -> Union[bool, "Literal[-1]"]:
        opts = self.get_options()
        if not opts:
            return False

        if "ip" in opts:
            try:
                self.ip_check()
            except (ValueError, ipaddress.AddressValueError) as e:
                logging.exception(e)
                return -1

        return True

    def setup_network(self) -> None:
        self.Config = Gio.Settings(schema_id="org.blueman.network")

        nap_enable = self._builder.get_widget("nap-enable", Gtk.CheckButton)
        r_dnsmasq = self._builder.get_widget("r_dnsmasq", Gtk.RadioButton)
        r_dhcpd = self._builder.get_widget("r_dhcpd", Gtk.RadioButton)
        r_udhcpd = self._builder.get_widget("r_udhcpd", Gtk.RadioButton)
        net_ip = self._builder.get_widget("net_ip", Gtk.Entry)
        rb_nm = self._builder.get_widget("rb_nm", Gtk.RadioButton)
        rb_blueman = self._builder.get_widget("rb_blueman", Gtk.RadioButton)
        rb_dun_nm = self._builder.get_widget("rb_dun_nm", Gtk.RadioButton)
        rb_dun_blueman = self._builder.get_widget("rb_dun_blueman", Gtk.RadioButton)

        nap_frame = self._builder.get_widget("nap_frame", Gtk.Frame)
        warning = self._builder.get_widget("warning", Gtk.Box)

        if not self.Config["nap-enable"]:
            nap_frame.props.sensitive = False

        if self.Config["ip4-address"] is not None:
            net_ip.props.text = self.Config["ip4-address"]
            nap_enable.props.active = True
        else:
            net_ip.props.text = f"10.{randint(0, 255):d}.{randint(0, 255):d}.1"

        if self.Config["dhcp-handler"] is None:
            nap_frame.props.sensitive = False
            nap_enable.props.active = False
            r_dnsmasq.props.active = True
            self.Config["nap-enable"] = False

        have_dhcpd = have("dhcpd3") or have("dhcpd")
        have_dnsmasq = have("dnsmasq")
        have_udhcpd = have("udhcpd")

        if self.Config["dhcp-handler"] == "DnsMasqHandler" and have_dnsmasq:
            r_dnsmasq.props.active = True
        elif self.Config["dhcp-handler"] == "DhcpdHandler" and have_dhcpd:
            r_dhcpd.props.active = True
        elif self.Config["dhcp-handler"] == "UdhcpdHandler" and have_udhcpd:
            r_udhcpd.props.active = True
        else:
            r_dnsmasq.props.active = True

        if not have_dnsmasq and not have_dhcpd and not have_udhcpd:
            nap_frame.props.sensitive = False
            warning.props.visible = True
            warning.props.sensitive = True
            nap_enable.props.sensitive = False
            self.Config["nap-enable"] = False

        if not have_dnsmasq:
            r_dnsmasq.props.sensitive = False
            r_dnsmasq.props.active = False

        if not have_dhcpd:
            r_dhcpd.props.sensitive = False
            r_dhcpd.props.active = False

        if not have_udhcpd:
            r_udhcpd.props.sensitive = False
            r_udhcpd.props.active = False

        r_dnsmasq.connect("toggled", lambda x: self.option_changed_notify("dnsmasq"))
        r_dhcpd.connect("toggled", lambda x: self.option_changed_notify("dhcpd"))
        r_udhcpd.connect("toggled", lambda x: self.option_changed_notify("udhcpd"))

        net_ip.connect("changed", lambda x: self.option_changed_notify("ip", False))
        nap_enable.connect("toggled", lambda x: self.option_changed_notify("nap_enable"))

        self.Config.bind("nap-enable", nap_enable, "active", Gio.SettingsBindFlags.GET)

        nap_enable.bind_property("active", nap_frame, "sensitive", GObject.BindingFlags.DEFAULT)

        applet = AppletService()

        avail_plugins = applet.QueryAvailablePlugins()
        active_plugins = applet.QueryPlugins()

        def dun_support_toggled(rb: Gtk.RadioButton, x: str) -> None:
            if rb.props.active and x == "nm":
                applet.SetPluginConfig('(sb)', "PPPSupport", False)
                applet.SetPluginConfig('(sb)', "NMDUNSupport", True)
            elif rb.props.active and x == "blueman":
                applet.SetPluginConfig('(sb)', "NMDUNSupport", False)
                applet.SetPluginConfig('(sb)', "PPPSupport", True)

        def pan_support_toggled(rb: Gtk.RadioButton, x: str) -> None:
            if rb.props.active and x == "nm":
                applet.SetPluginConfig('(sb)', "DhcpClient", False)
                applet.SetPluginConfig('(sb)', "NMPANSupport", True)

            elif rb.props.active and x == "blueman":
                applet.SetPluginConfig('(sb)', "NMPANSupport", False)
                applet.SetPluginConfig('(sb)', "DhcpClient", True)

        if "PPPSupport" in active_plugins:
            rb_dun_blueman.props.active = True

        if "NMDUNSupport" in avail_plugins:
            rb_dun_nm.props.sensitive = True
        else:
            rb_dun_nm.props.sensitive = False
            rb_dun_nm.props.tooltip_text = _("Not currently supported with this setup")

        if "DhcpClient" in active_plugins:
            rb_blueman.props.active = True

        if "NMPANSupport" in avail_plugins:
            rb_nm.props.sensitive = True
        else:
            rb_nm.props.sensitive = False
            rb_nm.props.tooltip_text = _("Not currently supported with this setup")

        if "NMPANSupport" in active_plugins:
            rb_nm.props.active = True

        if "NMDUNSupport" in active_plugins:
            rb_dun_nm.props.active = True

        rb_nm.connect("toggled", pan_support_toggled, "nm")
        rb_blueman.connect("toggled", pan_support_toggled, "blueman")

        rb_dun_nm.connect("toggled", dun_support_toggled, "nm")
        rb_dun_blueman.connect("toggled", dun_support_toggled, "blueman")



from gettext import gettext as _
from typing import Optional, TYPE_CHECKING, List, Any, cast, Callable, Set, Dict
import html
import logging
import cairo
import os

from blueman.bluez.Adapter import Adapter
from blueman.bluez.Battery import Battery
from blueman.bluez.Device import Device
from blueman.bluez.Manager import Manager
from blueman.gui.DeviceList import DeviceList
from blueman.DeviceClass import get_minor_class, get_major_class, gatt_appearance_to_name
from blueman.gui.GenericList import ListDataDict
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu
from blueman.Constants import PIXMAP_PATH
from blueman.Functions import launch
from blueman.Sdp import ServiceUUID, OBEX_OBJPUSH_SVCLASS_ID
from blueman.gui.GtkAnimation import TreeRowFade, CellFade, AnimBase
from _blueman import ConnInfoReadError, conn_info

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gio
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Pango

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class SurfaceObject(GObject.Object):
    __gtype_name__ = "SurfaceObject"

    def __init__(self, surface: cairo.ImageSurface) -> None:
        super().__init__()
        self.surface = surface


class ManagerDeviceList(DeviceList):
    def __init__(self, inst: "Blueman", adapter: Optional[str] = None) -> None:
        cr = Gtk.CellRendererText()
        cr.props.ellipsize = Pango.EllipsizeMode.END
        tabledata: List[ListDataDict] = [
            # device picture
            {"id": "device_surface", "type": SurfaceObject, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "celldata_func": (self._set_cell_data, None)},
            # device caption
            {"id": "caption", "type": str, "renderer": cr,
             "render_attrs": {"markup": 1}, "view_props": {"expand": True}},
            {"id": "battery_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "battery")},
            {"id": "rssi_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "rssi")},
            {"id": "tpl_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "tpl")},
            {"id": "alias", "type": str},  # used for quick access instead of device.GetProperties
            {"id": "connected", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "paired", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "trusted", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "objpush", "type": bool},  # used to set Send File button
            {"id": "battery", "type": float},
            {"id": "rssi", "type": float},
            {"id": "tpl", "type": float},
            {"id": "cell_fader", "type": CellFade},
            {"id": "row_fader", "type": TreeRowFade},
            {"id": "initial_anim", "type": bool},
            {"id": "blocked", "type": bool}
        ]
        super().__init__(adapter, tabledata)
        self.set_name("ManagerDeviceList")
        self.set_headers_visible(False)
        self.props.has_tooltip = True
        self.Blueman = inst

        self._monitored_devices: Set[str] = set()

        self.manager.connect_signal("battery-created", self.on_battery_created)
        self.manager.connect_signal("battery-removed", self.on_battery_removed)
        self._batteries: Dict[str, Battery] = {}

        self.Config = Gio.Settings(schema_id="org.blueman.general")
        self.Config.connect('changed', self._on_settings_changed)
        # Set the correct sorting
        self._on_settings_changed(self.Config, "sort-by")
        self._on_settings_changed(self.Config, "sort-type")

        self.connect("query-tooltip", self.tooltip_query)
        self.tooltip_row: Optional[Gtk.TreePath] = None
        self.tooltip_col: Optional[Gtk.TreeViewColumn] = None

        self.connect("popup-menu", self._on_popup_menu)
        self.connect("button-press-event", self._on_event_clicked)
        self.connect("button-release-event", self._on_event_clicked)
        self.connect("key-press-event", self._on_key_pressed)

        self.menu: Optional[ManagerDeviceMenu] = None

        self.connect("drag_data_received", self.drag_recv)
        self.connect("drag-motion", self.drag_motion)

        Gtk.Widget.drag_dest_set(self, Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT)
        Gtk.Widget.drag_dest_add_uri_targets(self)

        self.set_search_equal_func(self.search_func)
        self.filter.set_visible_func(self.filter_func)

    def _on_settings_changed(self, settings: Gio.Settings, key: str) -> None:
        if key in ('sort-by', 'sort-order'):
            sort_by = settings['sort-by']
            sort_order = settings['sort-order']

            if sort_order == 'ascending':
                sort_type = Gtk.SortType.ASCENDING
            else:
                sort_type = Gtk.SortType.DESCENDING

            column_id = self.ids.get(sort_by)

            if column_id:
                self.liststore.set_sort_column_id(column_id, sort_type)

    def on_icon_theme_changed(self, _icon_them: Gtk.IconTheme) -> None:
        for row in self.liststore:
            device = self.get(row.iter, "device")["device"]
            self.row_setup_event(row.iter, device)

    def on_battery_created(self, _manager: Manager, obj_path: str) -> None:
        if obj_path not in self._batteries:
            battery_proxy = Battery(obj_path=obj_path)
            self._batteries[obj_path] = battery_proxy
            logging.debug(f"{obj_path} {battery_proxy['Percentage']}")

    def on_battery_removed(self, _manager: Manager, obj_path: str) -> None:
        if obj_path in self._batteries:
            battery = self._batteries.pop(obj_path)
            battery.destroy()

    def search_func(self, model: Gtk.TreeModel, column: int, key: str, tree_iter: Gtk.TreeIter) -> bool:
        row = self.get(tree_iter, "caption")
        if key.lower() in row["caption"].lower():
            return False
        logging.info(f"{model} {column} {key} {tree_iter}")
        return True

    def filter_func(self, _model: Gtk.TreeModel, tree_iter: Gtk.TreeIter, _data: Any) -> bool:
        row = self.get(tree_iter, "no_name", "device")
        device = row["device"]
        klass = get_minor_class(device["Class"]) if device is not None else None

        if row["no_name"] and self.Config["hide-unnamed"] and klass not in (_("Keyboard"), _("Combo")):
            logging.info("Hiding unnamed device")
            return False
        else:
            return True

    def drag_recv(self, _widget: Gtk.Widget, context: Gdk.DragContext, x: int, y: int, selection: Gtk.SelectionData,
                  _info: int, time: int) -> None:

        uris = list(selection.get_uris())

        context.finish(True, False, time)

        path = self.get_path_at_pos(x, y)
        if path:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None
            device = self.get(tree_iter, "device")["device"]
            command = f"blueman-sendto --device={device['Address']}"

            launch(command, paths=uris, name=_("File Sender"))
            context.finish(True, False, time)
        else:
            context.finish(False, False, time)

    def drag_motion(self, _widget: Gtk.Widget, drag_context: Gdk.DragContext, x: int, y: int, timestamp: int) -> bool:
        result = self.get_path_at_pos(x, y)
        if result is not None:
            path = result[0]
            assert path is not None
            path = self.filter.convert_path_to_child_path(path)
            if path is None:
                return False

            if not self.selection.path_is_selected(path):
                tree_iter = self.get_iter(path)
                assert tree_iter is not None
                has_obj_push = self._has_objpush(self.get(tree_iter, "device")["device"])
                if has_obj_push:
                    Gdk.drag_status(drag_context, Gdk.DragAction.COPY, timestamp)
                    self.set_cursor(path)
                    return True
                else:
                    Gdk.drag_status(drag_context, Gdk.DragAction.DEFAULT, timestamp)
                    return False
            return False
        else:
            Gdk.drag_status(drag_context, Gdk.DragAction.DEFAULT, timestamp)
            return False

    def _on_popup_menu(self, _widget: Gtk.Widget) -> bool:
        if self.menu is None:
            self.menu = ManagerDeviceMenu(self.Blueman)

        window = self.get_window()
        assert window is not None
        selected = self.selected()
        assert selected is not None
        rect = self.get_cell_area(self.liststore.get_path(selected), self.get_column(1))
        self.menu.popup_at_rect(window, rect, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH)

        return True

    def _on_event_clicked(self, _widget: Gtk.Widget, event: Gdk.Event) -> bool:
        if event.type not in (Gdk.EventType._2BUTTON_PRESS, Gdk.EventType.BUTTON_PRESS):
            return False

        posdata = self.get_path_at_pos(int(cast(Gdk.EventButton, event).x), int(cast(Gdk.EventButton, event).y))
        if posdata is None:
            return False
        else:
            path = posdata[0]
            assert path is not None

        tree_iter = self.filter.get_iter(path)
        assert tree_iter is not None
        child_iter = self.filter.convert_iter_to_child_iter(tree_iter)
        assert child_iter is not None
        row = self.get(child_iter, "device", "connected")
        if not row:
            return False

        if self.menu is None:
            self.menu = ManagerDeviceMenu(self.Blueman)

        if event.type == Gdk.EventType._2BUTTON_PRESS and cast(Gdk.EventButton, event).button == 1:
            if self.menu.show_generic_connect_calc(row["device"]['UUIDs']):
                if row["connected"]:
                    self.menu.disconnect_service(row["device"])
                elif Adapter(obj_path=row["device"]["Adapter"])["Powered"]:
                    self.menu.connect_service(row["device"])

        if event.type == Gdk.EventType.BUTTON_PRESS and cast(Gdk.EventButton, event).button == 3:
            self.menu.popup_at_pointer(event)

        return False

    def _on_key_pressed(self, _widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
        if not (event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_c):
            return False

        selected = self.selected()
        if not selected:
            return False

        row = self.get(selected, "device")
        if not row:
            return False

        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(row["device"]["Address"], -1)
        return True

    def _load_surface(self, icon_name: str, size: int) -> cairo.ImageSurface:
        window = self.get_window()
        scale = self.get_scale_factor()
        icon_info = self.icon_theme.lookup_icon_for_scale(icon_name, size, scale, Gtk.IconLookupFlags.FORCE_SIZE)

        if icon_info is None:
            logging.error(f"Failed to look up icon \"{icon_name}\" likely due to broken icon theme.")
            missing_icon_info = self.icon_theme.lookup_icon_for_scale(
                "image-missing",
                size,
                scale,
                Gtk.IconLookupFlags.FORCE_SIZE
            )
            assert missing_icon_info is not None
            return cast(cairo.ImageSurface, missing_icon_info.load_surface(window))
        else:
            return cast(cairo.ImageSurface, icon_info.load_surface(window))

    def _make_device_icon(self, icon_name: str, is_paired: bool, is_connected: bool, is_trusted: bool,
                          is_blocked: bool) -> cairo.ImageSurface:
        scale = self.get_scale_factor()
        target = self._load_surface(icon_name, 48)
        ctx = cairo.Context(target)

        if is_connected or is_paired:
            icon = "blueman-connected-emblem" if is_connected else "blueman-paired-emblem"
            paired_surface = self._load_surface(icon, 16)
            ctx.set_source_surface(paired_surface, 1 / scale, 1 / scale)
            ctx.paint_with_alpha(0.8)

        if is_trusted:
            trusted_surface = self._load_surface("blueman-trusted-emblem", 16)
            assert isinstance(target, cairo.ImageSurface)
            height = target.get_height()
            mini_height = trusted_surface.get_height()
            y = height / scale - mini_height / scale - 1 / scale

            ctx.set_source_surface(trusted_surface, 1 / scale, y)
            ctx.paint_with_alpha(0.8)

        if is_blocked:
            blocked_surface = self._load_surface("blueman-blocked-emblem", 16)
            assert isinstance(target, cairo.ImageSurface)
            width = target.get_width()
            mini_width = blocked_surface.get_width()
            ctx.set_source_surface(blocked_surface, (width - mini_width - 1) / scale, 1 / scale)
            ctx.paint_with_alpha(0.8)

        return target

    def device_remove_event(self, object_path: str) -> None:
        tree_iter = self.find_device_by_path(object_path)
        assert tree_iter is not None

        iter_set, _child_tree_iter = self.filter.convert_child_iter_to_iter(tree_iter)
        if iter_set:
            row_fader = self.get(tree_iter, "row_fader")["row_fader"]
            self._prepare_fader(row_fader, lambda: self.__fader_finished(object_path))
            row_fader.animate(start=row_fader.get_state(), end=0.0, duration=400)

    def __fader_finished(self, object_path: str) -> None:
        super().device_remove_event(object_path)

    @staticmethod
    def make_caption(name: str, klass: str, address: str) -> str:
        return "<span size='x-large'>%(0)s</span>\n<span size='small'>%(1)s</span>\n<i>%(2)s</i>" \
               % {"0": html.escape(name), "1": klass, "2": address}

    @staticmethod
    def make_display_name(alias: str, klass: int, address: str) -> str:
        if alias.replace("-", ":") == address:
            return _("Unnamed device")
        else:
            return alias

    @staticmethod
    def get_device_class(device: Device) -> str:
        klass = get_minor_class(device['Class'])
        if klass != _("Uncategorized"):
            return klass
        else:
            return get_major_class(device['Class'])

    def row_setup_event(self, tree_iter: Gtk.TreeIter, device: Device) -> None:
        if not self.get(tree_iter, "initial_anim")["initial_anim"]:
            assert self.liststore is not None
            child_path = self.liststore.get_path(tree_iter)
            result = self.filter.convert_child_path_to_path(child_path)

            if child_path is not None:
                cell_fader = CellFade(self, child_path, [2, 3, 4, 5])
                row_fader = TreeRowFade(self, child_path)

                self.set(tree_iter, row_fader=row_fader, cell_fader=cell_fader)

                cell_fader.freeze()

                if result is not None:
                    self._prepare_fader(row_fader).animate(start=0.0, end=1.0, duration=500)
                    self.set(tree_iter, initial_anim=True)
                else:
                    self.set(tree_iter, initial_anim=False)

        has_objpush = self._has_objpush(device)
        klass = get_minor_class(device['Class'])
        # Bluetooth >= 4 devices use Appearance property
        appearance = device["Appearance"]
        if klass != _("Uncategorized") and klass != _("Unknown"):
            description = klass
        elif klass == _("Unknown") and appearance:
            description = gatt_appearance_to_name(appearance)
        else:
            description = get_major_class(device['Class'])

        surface = self._make_device_icon(device["Icon"], device["Paired"], device["Connected"], device["Trusted"],
                                         device["Blocked"])
        surface_object = SurfaceObject(surface)
        display_name = self.make_display_name(device.display_name, device["Class"], device['Address'])
        caption = self.make_caption(display_name, description, device['Address'])

        self.set(tree_iter, caption=caption, alias=display_name, objpush=has_objpush, device_surface=surface_object)

        try:
            self.row_update_event(tree_iter, "Trusted", device['Trusted'])
        except Exception as e:
            logging.exception(e)
        try:
            self.row_update_event(tree_iter, "Paired", device['Paired'])
        except Exception as e:
            logging.exception(e)
        try:
            self.row_update_event(tree_iter, "Connected", device["Connected"])
        except Exception as e:
            logging.exception(e)
        try:
            self.row_update_event(tree_iter, "Blocked", device["Blocked"])
        except Exception as e:
            logging.exception(e)

        if device["Connected"]:
            self._monitor_power_levels(tree_iter, device)

    def _monitor_power_levels(self, tree_iter: Gtk.TreeIter, device: Device) -> None:
        if device["Address"] in self._monitored_devices:
            return

        assert self.Adapter is not None
        cinfo = conn_info(device["Address"], os.path.basename(self.Adapter.get_object_path()))
        try:
            cinfo.init()
        except ConnInfoReadError:
            logging.warning("Failed to get power levels, probably a LE device.")

        model = self.liststore
        assert isinstance(model, Gtk.TreeModel)
        r = Gtk.TreeRowReference.new(model, model.get_path(tree_iter))
        self._update_power_levels(tree_iter, device, cinfo)
        GLib.timeout_add(1000, self._check_power_levels, r, cinfo, device["Address"])
        self._monitored_devices.add(device["Address"])

    def _check_power_levels(self, row_ref: Gtk.TreeRowReference, cinfo: conn_info, address: str) -> bool:
        if not row_ref.valid():
            logging.warning("stopping monitor (row does not exist)")
            cinfo.deinit()
            self._monitored_devices.remove(address)
            return False

        tree_iter = self.get_iter(row_ref.get_path())
        assert tree_iter is not None

        device = self.get(tree_iter, "device")["device"]

        if device["Connected"]:
            self._update_power_levels(tree_iter, device, cinfo)
            return True
        else:
            cinfo.deinit()
            self._disable_power_levels(tree_iter)
            self._monitored_devices.remove(address)
            return False

    def row_update_event(self, tree_iter: Gtk.TreeIter, key: str, value: Any) -> None:
        logging.info(f"{key} {value}")

        device = self.get(tree_iter, "device")["device"]

        if key in ("Blocked", "Connected", "Paired", "Trusted"):
            surface = self._make_device_icon(device["Icon"], device["Paired"], device["Connected"], device["Trusted"],
                                             device["Blocked"])
            self.set(tree_iter, device_surface=SurfaceObject(surface))

        if key == "Trusted":
            if value:
                self.set(tree_iter, trusted=True)
            else:
                self.set(tree_iter, trusted=False)

        elif key == "Paired":
            if value:
                self.set(tree_iter, paired=True)
            else:
                self.set(tree_iter, paired=False)

        elif key == "Alias":
            c = self.make_caption(value, self.get_device_class(device), device['Address'])
            name = self.make_display_name(device.display_name, device["Class"], device["Address"])
            self.set(tree_iter, caption=c, alias=name)

        elif key == "UUIDs":
            has_objpush = self._has_objpush(device)
            self.set(tree_iter, objpush=has_objpush)

        elif key == "Connected":
            self.set(tree_iter, connected=value)

            if value:
                self._monitor_power_levels(tree_iter, device)
            else:
                self._disable_power_levels(tree_iter)
        elif key == "Name":
            self.set(tree_iter, no_name=False)
            self.filter.refilter()

        elif key == "Blocked":
            self.set(tree_iter, blocked=value)

    def _update_power_levels(self, tree_iter: Gtk.TreeIter, device: Device, cinfo: conn_info) -> None:
        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "lq", "tpl")

        bars = {}

        obj_path = device.get_object_path()
        if obj_path in self._batteries:
            bars["battery"] = self._batteries[obj_path]["Percentage"]

        # cinfo init may fail for bluetooth devices version 4 and up
        # FIXME Workaround is horrible and we should show something better
        if cinfo.failed:
            bars.update({"rssi": 100.0, "tpl": 100.0})
        else:
            try:
                bars["rssi"] = max(50 + float(cinfo.get_rssi()) / 127 * 50, 10)
            except ConnInfoReadError:
                bars["rssi"] = 50
            try:
                bars["tpl"] = max(50 + float(cinfo.get_tpl()) / 127 * 50, 10)
            except ConnInfoReadError:
                bars["tpl"] = 50

        if row["battery"] == row["rssi"] == row["tpl"] == 0:
            self._prepare_fader(row["cell_fader"]).animate(start=0.0, end=1.0, duration=400)

        w = 14 * self.get_scale_factor()
        h = 48 * self.get_scale_factor()

        for (name, perc) in bars.items():
            if round(row[name], -1) != round(perc, -1):
                icon_name = f"blueman-{name}-{int(round(perc, -1))}.png"
                icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(os.path.join(PIXMAP_PATH, icon_name), w, h, True)
                self.set(tree_iter, **{name: perc, f"{name}_pb": icon})

    def _disable_power_levels(self, tree_iter: Gtk.TreeIter) -> None:
        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "tpl")
        if row["battery"] == row["rssi"] == row["tpl"] == 0:
            return

        self.set(tree_iter, battery=0, rssi=0, tpl=0)
        self._prepare_fader(row["cell_fader"], lambda: self.set(tree_iter, battery_pb=None, rssi_pb=None,
                                                                tpl_pb=None)).animate(start=1.0, end=0.0, duration=400)

    def _prepare_fader(self, fader: AnimBase, callback: Optional[Callable[[], None]] = None) -> AnimBase:
        def on_finished(finished_fader: AnimBase) -> None:
            finished_fader.disconnect(handler)
            finished_fader.freeze()
            if callback:
                callback()

        fader.thaw()
        handler = fader.connect("animation-finished", on_finished)
        return fader

    def tooltip_query(self, _tw: Gtk.Widget, x: int, y: int, _kb: bool, tooltip: Gtk.Tooltip) -> bool:
        path = self.get_path_at_pos(x, y)
        if path is None:
            return False

        if path[0] != self.tooltip_row or path[1] != self.tooltip_col:
            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return False

        if path[1] == self.columns["device_surface"]:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None

            row = self.get(tree_iter, "connected", "trusted", "paired", "blocked")
            str_list = []
            if row["connected"]:
                str_list.append(_("Connected"))
            if row["trusted"]:
                str_list.append(_("Trusted"))
            if row["paired"]:
                str_list.append(_("Paired"))
            if row["blocked"]:
                str_list.append(_("Blocked"))

            text = ", ".join(str_list)
            if text:
                tooltip.set_markup(f"<b>{text}</b>")
            else:
                return False

            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return True

        elif path[1] == self.columns["battery_pb"] \
                or path[1] == self.columns["tpl_pb"] \
                or path[1] == self.columns["rssi_pb"]:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None

            dt = self.get(tree_iter, "connected")["connected"]
            if not dt:
                return False

            lines = [_("<b>Connected</b>")]

            battery = self.get(tree_iter, "battery")["battery"]
            rssi = self.get(tree_iter, "rssi")["rssi"]
            tpl = self.get(tree_iter, "tpl")["tpl"]

            if battery != 0:
                if path[1] == self.columns["battery_pb"]:
                    lines.append(f"<b>Battery: {int(battery)}%</b>")
                else:
                    lines.append(f"Battery: {int(battery)}%")

            if rssi != 0:
                if rssi < 30:
                    rssi_state = _("Poor")
                elif rssi < 40:
                    rssi_state = _("Sub-optimal")
                elif rssi < 60:
                    rssi_state = _("Optimal")
                elif rssi < 70:
                    rssi_state = _("Much")
                else:
                    rssi_state = _("Too much")

                if path[1] == self.columns["rssi_pb"]:
                    lines.append(_("<b>Received Signal Strength: %(rssi)u%%</b> <i>(%(rssi_state)s)</i>") %
                                 {"rssi": rssi, "rssi_state": rssi_state})
                else:
                    lines.append(_("Received Signal Strength: %(rssi)u%% <i>(%(rssi_state)s)</i>") %
                                 {"rssi": rssi, "rssi_state": rssi_state})

            if tpl != 0:
                if tpl < 30:
                    tpl_state = _("Low")
                elif tpl < 40:
                    tpl_state = _("Sub-optimal")
                elif tpl < 60:
                    tpl_state = _("Optimal")
                elif tpl < 70:
                    tpl_state = _("High")
                else:
                    tpl_state = _("Very High")

                if path[1] == self.columns["tpl_pb"]:
                    lines.append(_("<b>Transmit Power Level: %(tpl)u%%</b> <i>(%(tpl_state)s)</i>") %
                                 {"tpl": tpl, "tpl_state": tpl_state})
                else:
                    lines.append(_("Transmit Power Level: %(tpl)u%% <i>(%(tpl_state)s)</i>") %
                                 {"tpl": tpl, "tpl_state": tpl_state})

            tooltip.set_markup("\n".join(lines))
            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return True
        return False

    def _has_objpush(self, device: Device) -> bool:
        if device is None:
            return False

        for uuid in device["UUIDs"]:
            if ServiceUUID(uuid).short_uuid == OBEX_OBJPUSH_SVCLASS_ID:
                return True
        return False

    def _set_cell_data(self, _col: Gtk.TreeViewColumn, cell: Gtk.CellRenderer, model: Gtk.TreeModelFilter,
                       tree_iter: Gtk.TreeIter, data: Optional[str]) -> None:
        tree_iter = model.convert_iter_to_child_iter(tree_iter)
        if data is None:
            row = self.get(tree_iter, "device_surface")
            cell.set_property("surface", row["device_surface"].surface)
        else:
            window = self.get_window()
            scale = self.get_scale_factor()
            pb = self.get(tree_iter, data + "_pb")[data + "_pb"]
            if pb:
                surface = Gdk.cairo_surface_create_from_pixbuf(pb, scale, window)
                cell.set_property("surface", surface)
            else:
                cell.set_property("surface", None)



from gettext import gettext as _
from operator import itemgetter
import html
import time
import logging
from typing import List, TYPE_CHECKING, Optional, Callable, cast, Union

from blueman.bluez.Device import Device
from blueman.bluez.errors import DBusNoSuchAdapterError
from blueman.gui.Notification import Notification
from blueman.Sdp import ServiceUUID
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.applet.Menu import MenuItem
from blueman.plugins.applet.PowerManager import PowerManager, PowerStateListener

if TYPE_CHECKING:
    from blueman.plugins.applet.Menu import Menu, SubmenuItemDict

    from typing_extensions import TypedDict

    class _ItemBase(TypedDict):
        adapter: str
        address: str
        alias: str
        icon: str
        name: str
        uuid: str

    class Item(_ItemBase):
        time: float
        device: str
        mitem: Optional[SubmenuItemDict]

    class StoredIcon(_ItemBase):
        time: str

REGISTRY_VERSION = 0


class RecentConns(AppletPlugin, PowerStateListener):
    __depends__ = ["DBusService", "Menu"]
    __icon__ = "document-open-recent-symbolic"
    __description__ = _("Provides a menu item that contains last used connections for quick access")
    __author__ = "Walmis"

    __gsettings__ = {
        "schema": "org.blueman.plugins.recentconns",
        "path": None
    }
    __options__ = {
        "max-items": {"type": int,
                      "default": 6,
                      # the maximum number of items RecentConns menu will display
                      "name": _("Maximum items"),
                      "desc": _("The maximum number of items recent connections menu will display."),
                      "range": (6, 20)},
        "recent-connections": {"type": list, "default": "[]"}
    }

    def on_load(self) -> None:
        self.__menuitems: List["SubmenuItemDict"] = []

        self._rebuild_menu()

    def on_power_state_changed(self, manager: PowerManager, state: bool) -> None:
        self._rebuild()

    def on_unload(self) -> None:
        self.parent.Plugins.Menu.unregister(self)

    def _rebuild(self) -> None:
        if 'PowerManager' in self.parent.Plugins.get_loaded() and \
                not self.parent.Plugins.PowerManager.get_bluetooth_status():
            for mitem in self._mitems:
                mitem.set_sensitive(False)
            return

        items = self._get_items()

        if len(items) == 0:
            for mitem in self._mitems:
                mitem.set_sensitive(False)
            return

        for mitem in self._mitems:
            mitem.set_sensitive(True)

        self.__menuitems = [self._build_menu_item(item) for item in items[:self.get_option("max-items")]]

        self._rebuild_menu()

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            self._rebuild()
        else:
            for mitem in self._mitems:
                mitem.set_sensitive(False)

    def on_device_created(self, path: str) -> None:
        self._rebuild()

    def on_device_removed(self, path: str) -> None:
        self._rebuild()

    def on_adapter_added(self, path: str) -> None:
        self._rebuild()

    def on_adapter_removed(self, path: str) -> None:
        self._rebuild()

    def notify(self, object_path: str, uuid: str) -> None:
        device = Device(obj_path=object_path)
        logging.info(f"{device} {uuid}")
        try:
            adapter = self.parent.Manager.get_adapter(device['Adapter'])
        except DBusNoSuchAdapterError:
            logging.warning("adapter not found")
            return

        item = {
            "adapter": adapter["Address"],
            "address": device['Address'],
            "alias": device.display_name,
            "icon": device['Icon'],
            "name": ServiceUUID(uuid).name,
            "uuid": uuid,
            "time": str(time.time()),
        }

        stored_items = self.get_option("recent-connections")

        for i in stored_items:
            if i["adapter"] == item["adapter"] and \
                    i["address"] == item["address"] and \
                    i["uuid"] == item["uuid"]:
                i["time"] = item["time"]
                i["device"] = object_path
                break
        else:
            stored_items.append(item)

        self.set_option("recent-connections", stored_items)

        self._rebuild()

    def on_item_activated(self, item: "Item") -> None:
        logging.info(f"Connect {item['address']} {item['uuid']}")

        assert item["mitem"] is not None

        item["mitem"]["sensitive"] = False
        self.parent.Plugins.Menu.on_menu_changed()

        def reply() -> None:
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            Notification(_("Connected"), _("Connected to %s") % item["mitem"]["text"],
                         icon_name=item["icon"]).show()
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        def err(reason: Union[Exception, str]) -> None:
            Notification(_("Failed to connect"), str(reason).split(": ")[-1],
                         icon_name="dialog-error").show()
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        self.parent.Plugins.DBusService.connect_service(item["device"], item["uuid"], reply, err)

    def _build_menu_item(self, item: "Item") -> "SubmenuItemDict":
        alias = html.escape(item["alias"])
        mitem: "SubmenuItemDict" = {
            "text": _("%(service)s on %(device)s") % {"service": item["name"], "device": alias},
            "markup": True,
            "icon_name": item["mitem"]["icon_name"] if item["mitem"] is not None else item["icon"],
            "sensitive": item["device"] is not None,
            "tooltip": None if item["device"] is None else _("Adapter for this connection is not available"),
            "callback": (item["mitem"]["callback"] if item["mitem"] is not None
                         else cast(Callable[[], None], lambda itm=item: self.on_item_activated(itm)))
        }

        item["mitem"] = mitem

        return mitem

    def _rebuild_menu(self) -> None:
        menu: "Menu" = self.parent.Plugins.Menu
        self._mitems: List[MenuItem] = []
        menu.unregister(self)
        menu.add(self, 52, text=_("Reconnect to…"), icon_name="document-open-recent-symbolic",
                 sensitive=False, callback=lambda: None)
        for (idx, item) in enumerate(self.__menuitems):
            self._mitems.append(menu.add(self, (53, idx), **item))
        menu.add(self, 59)

    def _get_device_path(self, adapter_path: str, address: str) -> Optional[str]:
        try:
            adapter = self.parent.Manager.get_adapter(adapter_path)
        except DBusNoSuchAdapterError:
            return None

        device = self.parent.Manager.find_device(address, adapter.get_object_path())
        return device.get_object_path() if device is not None else None

    def _get_items(self) -> List["Item"]:
        return sorted(
            ({
                "adapter": i["adapter"],
                "address": i["address"],
                "alias": i["alias"],
                "icon": i["icon"],
                "name": i["name"],
                "uuid": i["uuid"],
                "time": float(i["time"]),
                "device": device,
                "mitem": None
            }
                for i in self.get_option("recent-connections")
                if (device := self._get_device_path(i["adapter"], i["address"]))),
            key=itemgetter("time"),
            reverse=True
        )



from gettext import gettext as _
import logging
from typing import TYPE_CHECKING, Type, Dict, cast, Optional, TypeVar

from blueman.main.Builder import Builder
from blueman.main.PluginManager import PluginManager
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.BasePlugin import Option, BasePlugin

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk, Gdk, Gio, GLib, GObject

if TYPE_CHECKING:
    from blueman.main.Applet import BluemanApplet


class PluginItem(GObject.Object):
    __gtype_name__ = "PluginItem"

    class _Props:
        icon_name: str
        plugin_name: str
        description: str
        enabled: bool
        activatable: bool

    props: _Props

    icon_name = GObject.Property(type=str)
    plugin_name = GObject.Property(type=str)
    description = GObject.Property(type=str)
    enabled = GObject.Property(type=bool, default=False)
    activatable = GObject.Property(type=bool, default=False)

    def __init__(self, icon_name: str, plugin_name: str, description: str, enabled: bool, activatable: bool):
        super().__init__()
        self.props.icon_name = icon_name
        self.props.plugin_name = plugin_name
        self.props.description = description
        self.props.enabled = enabled
        self.props.activatable = activatable


class SettingsWidget(Gtk.Box):
    def __init__(self, inst: AppletPlugin, orientation: Gtk.Orientation = Gtk.Orientation.VERTICAL) -> None:
        super().__init__(
            name="SettingsWidget",
            orientation=orientation,
            spacing=2
        )
        self.inst = inst

        self.construct_settings()
        self.show_all()

    def construct_settings(self) -> None:
        for k, v in self.inst.__class__.__options__.items():
            if len(v) > 2:

                label = Gtk.Label(label=v["name"])
                label.props.xalign = 0.0

                w = self.get_control_widget(k, v)

                self.pack_start(w, False, False, 0)

                label = Gtk.Label(label="<i >" + v["desc"] + "</i>", wrap=True, use_markup=True, xalign=0.0)
                self.pack_start(label, False, False, 0)

    def handle_change(self, widget: Gtk.Widget, opt: str, params: "Option", prop: str) -> None:
        val = params["type"](getattr(widget.props, prop))
        logging.debug(f"changed {opt} {val}")

        self.inst.set_option(opt, val)

    def get_control_widget(self, opt: str, params: "Option") -> Gtk.Widget:
        if params["type"] == bool:
            c = Gtk.CheckButton(label=params["name"])

            c.props.active = self.inst.get_option(opt)
            c.connect("toggled", self.handle_change, opt, params, "active")
            return c

        elif params["type"] == int:
            b = Gtk.Box(spacing=6)
            label = Gtk.Label(label=params["name"])
            b.pack_start(label, False, False, 0)

            r = Gtk.SpinButton(numeric=True)
            b.pack_start(r, False, False, 6)

            r.set_increments(1, 3)
            r.set_range(params["range"][0], params["range"][1])

            r.props.value = self.inst.get_option(opt)
            r.connect("value-changed", self.handle_change, opt, params, "value")

            return b

        elif params["type"] == str:
            b = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=params["name"])
            b.pack_start(label, False, False, 0)

            e = Gtk.Entry()
            b.pack_start(e, False, False, 6)

            e.props.text = self.inst.get_option(opt)
            e.connect("changed", self.handle_change, opt, params, "text")

            return b
        raise ValueError()


class PluginDialog(Gtk.ApplicationWindow):
    def __init__(self, applet: "BluemanApplet") -> None:
        super().__init__(
            application=applet,
            title=_("Plugins"),
            icon_name="blueman",
            name="PluginDialog",
            border_width=10,
            default_width=490,
            default_height=380,
            resizable=False,
            visible=False
        )

        self.set_position(Gtk.WindowPosition.CENTER)

        self.applet = applet

        builder = Builder("applet-plugins-widget.ui")

        self.description = builder.get_widget("description", Gtk.Label)

        self.icon = builder.get_widget("icon", Gtk.Image)
        self.author_txt = builder.get_widget("author_txt", Gtk.Label)
        self.depends_txt = builder.get_widget("depends_txt", Gtk.Label)
        self.conflicts_txt = builder.get_widget("conflicts_txt", Gtk.Label)
        self.plugin_name = builder.get_widget("name", Gtk.Label)

        self.main_container = builder.get_widget("main_container", Gtk.Bin)
        self.content_grid = builder.get_widget("content", Gtk.Widget)

        self.b_prefs = builder.get_widget("b_prefs", Gtk.ToggleButton)
        self.b_prefs.connect("toggled", self.on_prefs_toggled)

        self.add(builder.get_widget("all", Gtk.Container))

        self.model = Gio.ListStore.new(PluginItem.__gtype__)
        self.listbox = builder.get_widget("plugin_listbox", Gtk.ListBox)
        self.listbox.bind_model(self.model, self._widget_factory)
        self.listbox.connect("row-selected", self._on_row_selected)

        plugin_list = builder.get_widget("plugin_list", Gtk.ScrolledWindow)
        plugin_info = builder.get_widget("main_scrolled_window", Gtk.ScrolledWindow)

        # Disable overlay scrolling
        if Gtk.get_minor_version() >= 16:
            plugin_list.props.overlay_scrolling = False
            plugin_info.props.overlay_scrolling = False

        self.populate()

        self.sig_a: int = self.applet.Plugins.connect("plugin-loaded", self.plugin_state_changed, True)
        self.sig_b: int = self.applet.Plugins.connect("plugin-unloaded", self.plugin_state_changed, False)
        self.connect("delete-event", self._on_close)

        close_action = Gio.SimpleAction.new("close", None)
        close_action.connect("activate", lambda x, y: self.close())

        self.add_action(close_action)

    def _add_plugin_action(self, name: str, state: bool, activatable: bool) -> None:
        logging.debug(f"adding action: {name}")
        action = Gio.SimpleAction.new_stateful(
            name, None, GLib.Variant.new_boolean(state)
        )
        action.set_property("enabled", activatable)
        self.add_action(action)
        action.connect("change-state", self._on_plugin_toggle)

    def _widget_factory(self, item: GObject.Object, _data: Optional[object] = None) -> Gtk.Widget:
        assert isinstance(item, PluginItem)
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5, visible=True)

        checkbutton = Gtk.CheckButton(visible=True, sensitive=item.props.activatable)
        box.add(checkbutton)
        checkbutton.set_action_name(f"win.{item.props.plugin_name}")

        self._add_plugin_action(item.props.plugin_name, item.props.enabled, item.props.activatable)
        # Set active after adding action
        checkbutton.set_active(item.props.enabled)

        plugin_im = Gtk.Image(icon_name=item.props.icon_name, visible=True)
        box.add(plugin_im)

        label = Gtk.Label(label=item.props.description, use_markup=True, visible=True)
        box.add(label)
        return box

    def _model_sort_func(self, item1: Optional[object], item2: Optional[object], _data: Optional[object] = None) -> int:
        assert isinstance(item1, PluginItem)
        assert isinstance(item2, PluginItem)

        if item1.props.activatable and not item2.props.activatable:
            return -1
        elif not item1.props.activatable and item2.props.activatable:
            return 1

        if item1.props.plugin_name < item2.props.plugin_name:
            return -1
        elif item1.props.plugin_name > item2.props.plugin_name:
            return 1

        return 0

    def _on_plugin_toggle(self, action: Gio.SimpleAction, state: GLib.Variant) -> None:
        action.set_state(state)
        plugin_name = action.get_name()

        deps = self.applet.Plugins.get_dependencies()[plugin_name]
        loaded = self.applet.Plugins.get_loaded()
        to_unload = [dep for dep in deps if dep in loaded]

        if to_unload:
            if not self._ask_unload(
                _("Plugin <b>\"%(0)s\"</b> depends on <b>%(1)s</b>. Unloading <b>%(1)s</b> will also unload <b>"
                  "\"%(0)s\"</b>.\nProceed?") % {"0": ", ".join(to_unload), "1": plugin_name}
            ):
                action.set_state(GLib.Variant.new_boolean(not state))
                return
        else:
            conflicts = self.applet.Plugins.get_conflicts()[plugin_name]
            to_unload = [conf for conf in conflicts if conf in loaded]

            if to_unload and not self._ask_unload(
                _("Plugin <b>%(0)s</b> conflicts with <b>%(1)s</b>. Loading <b>%(1)s</b> will unload <b>%(0)s</b>."
                  "\nProceed?") % {"0": ", ".join(to_unload), "1": plugin_name}
            ):
                action.set_state(GLib.Variant.new_boolean(not state))
                return

        for p in to_unload:
            logging.debug(f"unloading {p}")
            self.applet.Plugins.set_config(p, False)

        self.applet.Plugins.set_config(plugin_name, plugin_name not in self.applet.Plugins.get_loaded())

    def _on_row_selected(self, _lb: Gtk.ListBox, lbrow: Gtk.ListBoxRow) -> None:
        pos = lbrow.get_index()
        item = self.model.get_item(pos)
        assert isinstance(item, PluginItem)

        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[item.props.plugin_name]
        self.plugin_name.props.label = "<b>" + item.props.plugin_name + "</b>"
        self.icon.props.icon_name = cls.__icon__
        self.author_txt.props.label = cls.__author__
        self.description.props.label = cls.__description__

        if cls.__depends__:
            self.depends_txt.props.label = ", ".join(cls.__depends__)
        else:
            self.depends_txt.props.label = _("No dependencies")

        if cls.__conflicts__:
            self.conflicts_txt.props.label = ", ".join(cls.__conflicts__)
        else:
            self.conflicts_txt.props.label = _("No conflicts")

        if cls.is_configurable() and item.props.plugin_name in self.applet.Plugins.get_loaded():
            self.b_prefs.props.sensitive = True
        else:
            self.b_prefs.props.sensitive = False

        self.update_config_widget(cls)

    def _on_close(self, _widget: Gtk.Widget, _event: Gdk.Event) -> bool:
        self.applet.Plugins.disconnect(self.sig_a)
        self.applet.Plugins.disconnect(self.sig_b)
        return False

    def on_prefs_toggled(self, _button: Gtk.ToggleButton) -> None:
        row = self.listbox.get_selected_row()
        pos = row.get_index()
        item = cast(PluginItem, self.model.get_item(pos))
        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[item.props.plugin_name]

        self.update_config_widget(cls)

    def update_config_widget(self, cls: Type[AppletPlugin]) -> None:
        if self.b_prefs.props.active:
            if not cls.is_configurable():
                self.b_prefs.props.active = False
                return

            inst = cls.get_instance()
            if not inst:
                self.b_prefs.props.active = False
            else:
                c = self.main_container.get_child()
                assert c is not None
                self.main_container.remove(c)
                if isinstance(c, SettingsWidget):
                    c.destroy()

                self.main_container.add(SettingsWidget(inst))

        else:
            c = self.main_container.get_child()
            assert c is not None
            self.main_container.remove(c)
            if isinstance(c, SettingsWidget):
                c.destroy()
            self.main_container.add(self.content_grid)

    def populate(self) -> None:
        classes: Dict[str, Type[AppletPlugin]] = self.applet.Plugins.get_classes()
        loaded = self.applet.Plugins.get_loaded()
        for name, cls in classes.items():
            if cls.is_configurable():
                desc = f"<span weight=\"bold\">{name}</span>"
            else:
                desc = name
            plugin_item = PluginItem(cls.__icon__, name, desc, name in loaded, activatable=cls.__unloadable__)
            self.model.insert_sorted(plugin_item, self._model_sort_func)
        self.listbox.select_row(self.listbox.get_row_at_index(0))

    _T = TypeVar("_T", bound="BasePlugin")

    def plugin_state_changed(self, _plugins: PluginManager[_T], name: str, loaded: bool) -> None:
        logging.debug(f"{name} {loaded}")
        action = self.lookup_action(name)
        assert isinstance(action, Gio.SimpleAction)
        action.set_state(GLib.Variant.new_boolean(loaded))

        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[name]
        if not loaded:
            self.update_config_widget(cls)
            self.b_prefs.props.sensitive = False
        elif cls.is_configurable():
            self.b_prefs.props.sensitive = True

    def _ask_unload(self, text: str) -> bool:
        dialog = Gtk.MessageDialog(parent=self, type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
        dialog.props.secondary_use_markup = True
        dialog.props.icon_name = "blueman"
        dialog.props.text = _("Dependency issue")
        dialog.props.secondary_text = text

        resp = dialog.run()
        dialog.destroy()
        return resp == Gtk.ResponseType.YES



from ipaddress import IPv4Address
from typing import Callable
from unittest.mock import patch, Mock

from dbusmock import DBusTestCase
from gi.repository import GLib, Gio

from blueman.main.DNSServerProvider import DNSServerProvider
from test.testhelpers.DBusMock import DBusMock


@patch("blueman.main.DNSServerProvider.DNSServerProvider.RESOLVER_PATH", "/tmp/resolv.conf")
@patch(
    "blueman.main.DNSServerProvider.Gio.bus_get_sync",
    lambda bus_type: Gio.DBusConnection.new_for_address_sync(
        Gio.dbus_address_get_for_bus_sync(bus_type),
        Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION | Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT,
    )
)
class TestDNSServerProvider(DBusTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.start_system_bus()

    def test_resolver(self):
        with open("/tmp/resolv.conf", "w") as f:
            f.write("""# Test configuration
search mynet
nameserver 192.0.2.1
nameserver 2001:db8::1
nameserver 198.51.100.1""")

        self.assertListEqual(
            DNSServerProvider.get_servers(),
            [IPv4Address("192.0.2.1"), IPv4Address("198.51.100.1")]
        )

    def test_systemd_resolved(self):
        with DBusMock("org.freedesktop.resolve1", "/org/freedesktop/resolve1", Gio.BusType.SYSTEM) as mock:
            mock.add_property("org.freedesktop.resolve1.Manager", "DNS",
                              GLib.Variant("a(iiay)", [
                                  (0, 2, [203, 0, 113, 1]),
                                  (0, 10, [0x20, 0x01, 0x0d, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x01]),
                                  (3, 2, [198, 51, 100, 32]),
                                  (3, 10, [0x20, 0x01, 0x0d, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x02]),
                                  (5, 2, [198, 51, 100, 64]),
                                  (5, 10, [0x20, 0x01, 0x0d, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x03]),
                              ]))
            mock.add_method("org.freedesktop.resolve1.Manager", "GetLink", "i", "s",
                            "ret = f'/org/freedesktop/resolve1/link/_3{args[0]}'")
            mock.add_object("/org/freedesktop/resolve1/link/_33")\
                .add_property("org.freedesktop.resolve1.Link", "DefaultRoute", GLib.Variant("b", True))
            mock.add_object("/org/freedesktop/resolve1/link/_35")\
                .add_property("org.freedesktop.resolve1.Link", "DefaultRoute", GLib.Variant("b", False))

            self.assertListEqual(
                sorted(DNSServerProvider.get_servers()),
                [IPv4Address("198.51.100.32"), IPv4Address("203.0.113.1")]
            )

    def test_resolver_changed(self):
        self._test_changed(lambda: open("/tmp/resolv.conf", "w").close())

    def test_resolved_changed(self):
        def trigger():
            with DBusMock("org.freedesktop.resolve1", "/org/freedesktop/resolve1", Gio.BusType.SYSTEM) as dbus_mock:
                dbus_mock.add_property("org.freedesktop.resolve1.Manager", "DNS", GLib.Variant("a(iiay)", []))
                dbus_mock.set_property("org.freedesktop.resolve1.Manager", "DNS", GLib.Variant("a(iiay)", []))

        self._test_changed(trigger)

    @staticmethod
    def _test_changed(action: Callable[[], None]) -> None:
        mock = Mock()
        provider = DNSServerProvider()
        provider.connect("changed", mock)

        action()

        context = GLib.MainContext.default()
        while context.pending():
            context.iteration()
        mock.assert_called_with(provider)



import errno
import os
import ipaddress
import socket
from tempfile import mkstemp
from time import sleep
import logging
import signal
from typing import List, Tuple, Optional, Type

from blueman.Constants import DHCP_CONFIG_FILE
from blueman.Functions import have
from _blueman import create_bridge, destroy_bridge, BridgeException
from subprocess import call, Popen, PIPE

from blueman.main.DNSServerProvider import DNSServerProvider


class NetworkSetupError(Exception):
    pass


def _is_running(name: str, pid: int) -> bool:
    if not os.path.exists(f"/proc/{pid}"):
        return False

    with open(f"/proc/{pid}/cmdline") as f:
        return name in f.readline().replace("\0", " ")


def _read_pid_file(fname: str) -> Optional[int]:
    try:
        with open(fname) as f:
            return int(f.read())
    except (OSError, ValueError):
        return None


def _get_binary(*names: str) -> str:
    for name in names:
        path = have(name)
        if path:
            return path
    raise FileNotFoundError(f"{' '.join(names)} not found")


class DHCPHandler:
    _BINARIES: List[str]

    @property
    def _key(self) -> str:
        return self._BINARIES[-1]

    def __init__(self) -> None:
        self._pid: Optional[int] = None

    @staticmethod
    def _get_arguments(ip4_address: str) -> List[str]:
        return []

    @property
    def _pid_path(self) -> str:
        return f"/var/run/{self._key}.pan1.pid"

    def apply(self, ip4_address: str, ip4_mask: str) -> None:
        error = self._start(_get_binary(*self._BINARIES), ip4_address, ip4_mask,
                            [ip4_address if addr.is_loopback else str(addr)
                             for addr in DNSServerProvider.get_servers()])
        if error is None:
            logging.info(f"{self._key} started correctly")
            with open(self._pid_path) as f:
                self._pid = int(f.read())
            logging.info(f"pid {self._pid}")
            NetConf.lock("dhcp")
        else:
            error_msg = error.decode("UTF-8").strip()
            logging.info(error_msg)
            raise NetworkSetupError(f"{self._key} failed to start: {error_msg}")

    def _start(self, binary: str, ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> Optional[bytes]:
        ...

    def clean_up(self) -> None:
        self._clean_up_configuration()

        if NetConf.locked("dhcp"):
            if not self._pid:
                pid = _read_pid_file(self._pid_path)
            else:
                pid = self._pid

            if pid is not None:
                running_binary: Optional[str] = next(binary for binary in self._BINARIES if _is_running(binary, pid))
                if running_binary is not None:
                    print('Terminating ' + running_binary)
                    os.kill(pid, signal.SIGTERM)
            else:
                running_binary = None

            if pid is None or running_binary is None:
                logging.info("Stale dhcp lockfile found")

            NetConf.unlock("dhcp")

    def _clean_up_configuration(self) -> None:
        ...


class DnsMasqHandler(DHCPHandler):
    _BINARIES = ["dnsmasq"]

    def _start(self, binary: str, ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> Optional[bytes]:
        ipiface = ipaddress.ip_interface('/'.join((ip4_address, ip4_mask)))
        cmd = [binary, f"--pid-file={self._pid_path}", "--except-interface=lo",
               "--interface=pan1", "--bind-interfaces",
               f"--dhcp-range={ipiface.network[2]},{ipiface.network[-2]},60m",
               f"--dhcp-option=option:router,{ip4_address}"]

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(("localhost", 53)) == 0:
                cmd += ["--port=0", f"--dhcp-option=option:dns-server,{', '.join(dns_servers)}"]

        logging.info(cmd)
        p = Popen(cmd, stderr=PIPE)

        error = p.communicate()[1]

        return error if error else None


DHCPDSUBNET = '''#### BLUEMAN AUTOMAGIC SUBNET ####
# Everything inside this section is destroyed after config change
subnet %(ip_mask)s netmask %(netmask)s {
  option domain-name-servers %(dns)s;
  option subnet-mask %(netmask)s;
  option routers %(rtr)s;
  range %(start)s %(end)s;
}
#### END BLUEMAN AUTOMAGIC SUBNET ####'''


class DhcpdHandler(DHCPHandler):
    _BINARIES = ["dhcpd3", "dhcpd"]

    @staticmethod
    def _read_dhcp_config() -> Tuple[str, str]:
        dhcp_config = ''
        existing_subnet = ''
        start = end = False

        with open(DHCP_CONFIG_FILE) as f:
            for line in f:
                if line == '#### BLUEMAN AUTOMAGIC SUBNET ####\n':
                    start = True
                elif line == '#### END BLUEMAN AUTOMAGIC SUBNET ####\n':
                    if not start:
                        # Because of bug end string got left upon removal
                        continue
                    end = True

                if start and not end:
                    existing_subnet += line
                elif start and end:
                    existing_subnet += line
                    start = end = False
                else:
                    dhcp_config += line

        return dhcp_config, existing_subnet

    @staticmethod
    def _generate_subnet_config(ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> str:
        ipiface = ipaddress.ip_interface('/'.join((ip4_address, ip4_mask)))

        return DHCPDSUBNET % {"ip_mask": ipiface.network.network_address,
                              "netmask": ipiface.netmask,
                              "dns": ', '.join(dns_servers),
                              "rtr": ipiface.ip,
                              "start": ipiface.network[2],
                              "end": ipiface.network[-2]}

    def _start(self, binary: str, ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> Optional[bytes]:
        dhcp_config, existing_subnet = self._read_dhcp_config()

        subnet = self._generate_subnet_config(ip4_address, ip4_mask, dns_servers)

        with open(DHCP_CONFIG_FILE, "w") as f:
            f.write(dhcp_config)
            f.write(subnet)

        cmd = [binary, "-pf", self._pid_path, "pan1"]
        p = Popen(cmd, stderr=PIPE)

        error = p.communicate()[1]

        return None if p.returncode == 0 else error

    def _clean_up_configuration(self) -> None:
        dhcp_config, existing_subnet = self._read_dhcp_config()
        with open(DHCP_CONFIG_FILE, "w") as f:
            f.write(dhcp_config)


UDHCP_CONF_TEMPLATE = """start %(start)s
end %(end)s
interface pan1
pidfile %(pid_path)s
option subnet %(ip_mask)s
option dns %(dns)s
option router %(rtr)s
"""


class UdhcpdHandler(DHCPHandler):
    _BINARIES = ["udhcpd"]

    def _generate_config(self, ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> str:
        ipiface = ipaddress.ip_interface('/'.join((ip4_address, ip4_mask)))

        return UDHCP_CONF_TEMPLATE % {"ip_mask": ipiface.network.network_address,
                                      "dns": ', '.join(dns_servers),
                                      "rtr": ipiface.ip,
                                      "start": ipiface.network[2],
                                      "end": ipiface.network[-2],
                                      "pid_path": self._pid_path}

    def _start(self, binary: str, ip4_address: str, ip4_mask: str, dns_servers: List[str]) -> Optional[bytes]:
        config_file, self._config_path = mkstemp(prefix="udhcpd-")
        with open(config_file, "w", encoding="utf8") as f:
            f.write(self._generate_config(ip4_address, ip4_mask, dns_servers))

        logging.info(f"Running udhcpd with config file {self._config_path}")
        cmd = [binary, "-S", self._config_path]
        p = Popen(cmd, stderr=PIPE)
        error = p.communicate()[1]

        # udhcpd takes time to create pid file
        sleep(0.1)

        pid = _read_pid_file(self._pid_path)

        return None if p.pid and pid is not None and _is_running("udhcpd", pid) else error

    def _clean_up_configuration(self) -> None:
        if os.path.exists(self._config_path):
            os.remove(self._config_path)


class NetConf:
    _dhcp_handler: Optional[DHCPHandler] = None
    _ipt_rules: List[Tuple[str, str, str]] = []

    _IPV4_SYS_PATH = "/proc/sys/net/ipv4"
    _RUN_PATH = "/var/run"

    @classmethod
    def _enable_ip4_forwarding(cls) -> None:
        with open(f"{cls._IPV4_SYS_PATH}/ip_forward", "w") as f:
            f.write("1")

        for d in os.listdir(f"{cls._IPV4_SYS_PATH}/conf"):
            with open(f"{cls._IPV4_SYS_PATH}/conf/{d}/forwarding", "w") as f:
                f.write("1")

    @classmethod
    def _add_ipt_rule(cls, table: str, chain: str, rule: str) -> None:
        cls._ipt_rules.append((table, chain, rule))
        args = ["/sbin/iptables", "-t", table, "-A", chain] + rule.split(" ")
        logging.debug(" ".join(args))
        ret = call(args)
        logging.info(f"Return code {ret}")

    @classmethod
    def _del_ipt_rules(cls) -> None:
        for table, chain, rule in cls._ipt_rules:
            call(["/sbin/iptables", "-t", table, "-D", chain] + rule.split(" "))
        cls._ipt_rules = []
        cls.unlock("iptables")

    @classmethod
    def apply_settings(cls, ip4_address: str, ip4_mask: str, handler: Type["DHCPHandler"],
                       address_changed: bool) -> None:
        if not isinstance(cls._dhcp_handler, handler):
            if cls._dhcp_handler is not None:
                cls._dhcp_handler.clean_up()

            cls._dhcp_handler = handler()

        try:
            create_bridge("pan1")
        except BridgeException as e:
            if e.errno != errno.EEXIST:
                raise

        if address_changed or not cls.locked("netconfig"):
            cls._enable_ip4_forwarding()

            if have("ip"):
                ret = call(["ip", "link", "set", "dev", "pan1", "up"])
                if ret != 0:
                    raise NetworkSetupError("Failed to bring up interface pan1")

                ret = call(["ip", "address", "add", "/".join((ip4_address, ip4_mask)), "dev", "pan1"])
                if ret != 0:
                    raise NetworkSetupError(f"Failed to add ip address {ip4_address}"
                                            f"with netmask {ip4_mask}")
            elif have('ifconfig'):
                ret = call(["ifconfig", "pan1", ip4_address, "netmask", ip4_mask, "up"])
                if ret != 0:
                    raise NetworkSetupError(f"Failed to add ip address {ip4_address}"
                                            f"with netmask {ip4_mask}")
            else:
                raise NetworkSetupError(
                    "Neither ifconfig or ip commands are found. Please install net-tools or iproute2")

            cls.lock("netconfig")

        if address_changed or not cls.locked("iptables"):
            cls._del_ipt_rules()

            cls._add_ipt_rule("nat", "POSTROUTING", f"-s {ip4_address}/{ip4_mask} -j MASQUERADE")
            cls._add_ipt_rule("filter", "FORWARD", "-i pan1 -j ACCEPT")
            cls._add_ipt_rule("filter", "FORWARD", "-o pan1 -j ACCEPT")
            cls._add_ipt_rule("filter", "FORWARD", "-i pan1 -j ACCEPT")
            cls.lock("iptables")

        if address_changed or not NetConf.locked("dhcp"):
            cls._dhcp_handler.clean_up()
            cls._dhcp_handler.apply(ip4_address, ip4_mask)

    @classmethod
    def clean_up(cls) -> None:
        logging.info(cls)

        if cls._dhcp_handler:
            cls._dhcp_handler.clean_up()

        try:
            destroy_bridge("pan1")
        except BridgeException:
            pass
        cls.unlock("netconfig")

        cls._del_ipt_rules()

    @classmethod
    def lock(cls, key: str) -> None:
        with open(f"{cls._RUN_PATH}/blueman-{key}", "w"):
            pass

    @classmethod
    def unlock(cls, key: str) -> None:
        try:
            os.unlink(f"{cls._RUN_PATH}/blueman-{key}")
        except OSError:
            pass

    @classmethod
    def locked(cls, key: str) -> bool:
        return os.path.exists(f"{cls._RUN_PATH}/blueman-{key}")



from gettext import gettext as _
import logging
from typing import Optional, Any, List, Set

from gi.repository import GLib

from blueman.bluez.Battery import Battery
from blueman.bluez.Device import Device
from blueman.bluez.errors import BluezDBusException
from blueman.main.BatteryWatcher import BatteryWatcher
from blueman.plugins.AppletPlugin import AppletPlugin

from blueman.plugins.applet.StatusIcon import StatusIconProvider
from blueman.main.PluginManager import PluginManager


class ShowConnected(AppletPlugin, StatusIconProvider):
    __author__ = "Walmis"
    __depends__ = ["StatusIcon"]
    __icon__ = "bluetooth-symbolic"
    __description__ = _("Adds an indication on the status icon when Bluetooth is active and shows the "
                        "connections in the tooltip.")

    def on_load(self) -> None:
        self._connections: Set[str] = set()
        self.active = False
        self.initialized = False
        self._handlers: List[int] = []
        self._handlers.append(self.parent.Plugins.connect('plugin-loaded', self._on_plugins_changed))
        self._handlers.append(self.parent.Plugins.connect('plugin-unloaded', self._on_plugins_changed))
        self._battery_watcher = BatteryWatcher(lambda *args: self.update_statusicon())

    def on_unload(self) -> None:
        self.parent.Plugins.StatusIcon.set_tooltip_text(None)
        self._connections = set()
        self.parent.Plugins.StatusIcon.icon_should_change()
        for handler in self._handlers:
            self.parent.Plugins.disconnect(handler)
        self._handlers = []
        del self._battery_watcher

    def on_status_icon_query_icon(self) -> Optional[str]:
        if self._connections:
            self.active = True
            return "blueman-active"
        else:
            self.active = False
            return None

    def enumerate_connections(self) -> bool:
        self._connections = {device.get_object_path()
                             for device in self.parent.Manager.get_devices()
                             if device["Connected"]}

        logging.info(f"Found {len(self._connections):d} existing connections")
        if (self._connections and not self.active) or (not self._connections and self.active):
            self.parent.Plugins.StatusIcon.icon_should_change()

        self.update_statusicon()

        return False

    def update_statusicon(self) -> None:
        if self._connections:
            def build_line(obj_path: str) -> str:
                line: str = Device(obj_path=obj_path)["Alias"]
                try:
                    return f"{line} 🔋{Battery(obj_path=obj_path)['Percentage']}%"
                except BluezDBusException:
                    return line

            self.parent.Plugins.StatusIcon.set_tooltip_title(_("Bluetooth Active"))
            self.parent.Plugins.StatusIcon.set_tooltip_text("\n".join(map(build_line, self._connections)))
        else:
            self.parent.Plugins.StatusIcon.set_tooltip_text(None)
            if 'PowerManager' in self.parent.Plugins.get_loaded():
                status = self.parent.Plugins.PowerManager.get_bluetooth_status()
                if status:
                    self.parent.Plugins.StatusIcon.set_tooltip_title(_("Bluetooth Enabled"))
                else:
                    self.parent.Plugins.StatusIcon.set_tooltip_title(_("Bluetooth Disabled"))
            else:
                self.parent.Plugins.StatusIcon.set_tooltip_title("Blueman")

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            if not self.initialized:
                GLib.timeout_add(0, self.enumerate_connections)
                self.initialized = True
            else:
                GLib.timeout_add(1000, self.enumerate_connections)
        else:
            self._connections = set()
            self.update_statusicon()

    def on_device_property_changed(self, path: str, key: str, value: Any) -> None:
        if key == "Connected":
            if value:
                self._connections.add(path)
            else:
                self._connections.remove(path)

            if (self._connections and not self.active) or (not self._connections and self.active):
                self.parent.Plugins.StatusIcon.icon_should_change()

            self.update_statusicon()

    def on_adapter_added(self, _path: str) -> None:
        self.enumerate_connections()

    def on_adapter_removed(self, _path: str) -> None:
        self.enumerate_connections()

    def _on_plugins_changed(self, _pluginmngr: PluginManager[AppletPlugin], name: str) -> None:
        if name == "PowerManager":
            self.update_statusicon()



from dbusmock import DBusTestCase
from gi.repository import Gio, GLib


class DBusMockObject:
    def __init__(self, name: str, path: str, system_bus: bool) -> None:
        self._proxy = Gio.DBusProxy(
            g_name=name,
            g_interface_name="org.freedesktop.DBus.Mock",
            g_object_path=path,
            g_connection=Gio.bus_get_sync(Gio.BusType.SYSTEM if system_bus else Gio.BusType.SESSION),
        )
        self._proxy.init()

    def add_method(self, interface: str, name: str, in_sig: str, out_sig: str, script: str) -> None:
        self._proxy.call_sync("AddMethod", GLib.Variant("(sssss)", (interface, name, in_sig, out_sig, script)),
                              Gio.DBusCallFlags.NONE, GLib.MAXINT)

    def add_property(self, interface: str, name: str, value: GLib.Variant) -> None:
        self._proxy.call_sync("AddProperty", GLib.Variant("(ssv)", (interface, name, value)),
                              Gio.DBusCallFlags.NONE, GLib.MAXINT)

    def set_property(self, interface: str, name: str, value: GLib.Variant) -> None:
        self._proxy.call_sync("org.freedesktop.DBus.Properties.Set", GLib.Variant("(ssv)", (interface, name, value)),
                              Gio.DBusCallFlags.NONE, GLib.MAXINT)


class DBusMock(DBusMockObject):
    def __init__(self, name: str, path: str, system_bus: bool) -> None:
        self._system_bus = system_bus
        self._process = DBusTestCase.spawn_server(name, path, "org.freedesktop.DBus.Mock", system_bus=system_bus)
        super().__init__(name, path, system_bus)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._process.terminate()
        self._process.wait()

    def add_object(self, path: str) -> DBusMockObject:
        self._proxy.call_sync(
            "AddObject", GLib.Variant("(ssa{sv}a(ssss))", (path, "org.freedesktop.DBus.Mock", {}, [])),
            Gio.DBusCallFlags.NONE, GLib.MAXINT)
        return DBusMockObject(self._proxy.props.g_name, path, self._system_bus)



from gettext import gettext as _
import logging

major_cls = (
    # translators: device class
    _("Miscellaneous"),
    # translators: device class
    _("Computer"),
    # translators: device class
    _("Phone"),
    # translators: device class
    _("Access point"),
    # translators: device class
    _("Audio/video"),
    # translators: device class
    _("Peripheral"),
    # translators: device class
    _("Imaging"),
    # translators: device class
    _("Wearable"),
    # translators: device class
    _("Toy"),
    # translators: device class
    _("Uncategorized")
)

computer_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Desktop"),
    # translators: device class
    _("Server"),
    # translators: device class
    _("Laptop"),
    # translators: device class
    _("Handheld"),
    # translators: device class
    _("Palm"),
    # translators: device class
    _("Wearable")
)

phone_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Cellular"),
    # translators: device class
    _("Cordless"),
    # translators: device class
    _("Smartphone"),
    # translators: device class
    _("Modem"),
    # translators: device class
    _("ISDN")
)

access_point_minor_cls = (
    # translators: device class
    _("Fully"),
    # translators: device class
    _("1–17 percent"),
    # translators: device class
    _("17–33 percent"),
    # translators: device class
    _("33–50 percent"),
    # translators: device class
    _("50–67 percent"),
    # translators: device class
    _("67–83 percent"),
    # translators: device class
    _("83–99 percent"),
    # translators: device class
    _("Not available")
)

audio_video_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Headset"),
    # translators: device class
    _("Handsfree"),
    # translators: device class
    _("Unknown"),
    # translators: device class
    _("Microphone"),
    # translators: device class
    _("Loudspeaker"),
    # translators: device class
    _("Headphones"),
    # translators: device class
    _("Portable audio"),
    # translators: device class
    _("Car audio"),
    # translators: device class
    _("Set-top box"),
    # translators: device class
    _("Hi-Fi audio"),
    # translators: device class
    _("VCR"),
    # translators: device class
    _("Video camera"),
    # translators: device class
    _("Camcorder"),
    # translators: device class
    _("Video monitor"),
    # translators: device class
    _("Video display and loudspeaker"),
    # translators: device class
    _("Video conferencing"),
    # translators: device class
    _("Unknown"),
    # translators: device class
    _("Gaming/Toy")
)

peripheral_minor_cls = (
    # translators: device class
    _("Uncategorized"),
    # translators: device class
    _("Keyboard"),
    # translators: device class
    _("Pointing"),
    # translators: device class
    _("Combo")
)

imaging_minor_cls = (
    # translators: device class
    _("Display"),
    # translators: device class
    _("Camera"),
    # translators: device class
    _("Scanner"),
    # translators: device class
    _("Printer")
)

wearable_minor_cls = (
    # translators: device class
    _("Wrist watch"),
    # translators: device class
    _("Pager"),
    # translators: device class
    _("Jacket"),
    # translators: device class
    _("Helmet"),
    # translators: device class
    _("Glasses")
)

toy_minor_cls = (
    # translators: device class
    _("Robot"),
    # translators: device class
    _("Vehicle"),
    # translators: device class
    _("Doll"),
    # translators: device class
    _("Controller"),
    # translators: device class
    _("Game")
)

gatt_appearance_categories = {
    0x00: ("Unknown", {0x00: _("Generic Unknown")}),
    0x01: ("Phone", {0x00: _("Generic Phone")}),
    0x02: ("Computer", {0x00: _("Generic Computer"),
                        0x01: _("Desktop Workstation"),
                        0x02: _("Server-class Computer"),
                        0x03: _("Laptop"),
                        0x04: _("Handheld PC/PDA (clamshell)"),
                        0x05: _("Palm-size PC/PDA"),
                        0x06: _("Wearable computer (watch size)"),
                        0x07: _("Tablet"),
                        0x08: _("Docking Station"),
                        0x09: _("All in One"),
                        0x0A: _("Blade Server"),
                        0x0B: _("Convertible"),
                        0x0C: _("Detachable"),
                        0x0D: _("IoT Gateway"),
                        0x0E: _("Mini PC"),
                        0x0F: _("Stick PC")}),
    0x03: ("Watch", {0x00: _("Generic Watch"),
                     0x01: _("Sports Watch"),
                     0x02: _("Smartwatch")}),
    0x04: ("Clock", {0x00: _("Generic Clock")}),
    0x05: ("Display", {0x00: _("Generic Display")}),
    0x06: ("Remote Control", {0x00: _("Generic Remote Control")}),
    0x07: ("Eye-glasses", {0x00: _("Generic Eye-glasses")}),
    0x08: ("Tag", {0x00: _("Generic Tag")}),
    0x09: ("Keyring", {0x00: _("Generic Keyring")}),
    0x0A: ("Media Player", {0x00: _("Generic Media Player")}),
    0x0B: ("Barcode Scanner", {0x00: _("Generic Barcode Scanner")}),
    0x0C: ("Thermometer", {0x00: _("Generic Thermometer"),
                           1: _("Ear Thermometer")}),
    0x0D: ("Heart Rate Sensor", {0x00: _("Generic Heart Rate Sensor"),
                                 1: _("Heart Rate Belt")}),
    0x0E: ("Blood Pressure", {0x00: _("Generic Blood Pressure"),
                              1: _("Arm Blood Pressure"),
                              2: _("Wrist Blood Pressure")}),
    0x0F: ("Human Interface Device", {0x00: _("Generic Human Interface Device"),
                                      0x01: _("Keyboard"),
                                      0x02: _("Mouse"),
                                      0x03: _("Joystick"),
                                      0x04: _("Gamepad"),
                                      0x05: _("Digitizer Tablet"),
                                      0x06: _("Card Reader"),
                                      0x07: _("Digital Pen"),
                                      0x08: _("Barcode Scanner"),
                                      0x09: _("Touchpad"),
                                      0x0a: _("Presentation Remote")}),
    0x10: ("Glucose Meter", {0x00: _("Generic Glucose Meter")}),
    0x11: ("Running Walking Sensor", {0x00: _("Generic Running Walking Sensor"),
                                      0x01: _("In-Shoe Running Walking Sensor"),
                                      0x02: _("On-Shoe Running Walking Sensor"),
                                      0x03: _("On-Hip Running Walking Sensor")}),
    0x12: ("Cycling", {0x00: _("Generic Cycling"),
                       0x01: _("Cycling Computer"),
                       0x02: _("Speed Sensor"),
                       0x03: _("Cadence Sensor"),
                       0x04: _("Power Sensor"),
                       0x05: _("Speed and Cadence Sensor")}),
    0x13: ("Control Device", {0x00: _("Generic Control Device"),
                              0x01: _("Switch"),
                              0x02: _("Multi-switch"),
                              0x03: _("Button"),
                              0x04: _("Slider"),
                              0x05: _("Rotary Switch"),
                              0x06: _("Touch Panel"),
                              0x07: _("Single Switch"),
                              0x08: _("Double Switch"),
                              0x09: _("Triple Switch"),
                              0x0A: _("Battery Switch"),
                              0x0B: _("Energy Harvesting Switch"),
                              0x0C: _("Push Button"),
                              0x0D: _("Dial")}),
    0x14: ("Network Device", {0x00: _("Generic Network Device"),
                              0x01: _("Access Point"),
                              0x02: _("Mesh Device"),
                              0x03: _("Mesh Network Proxy")}),
    0x15: ("Sensor", {0x00: _("Generic Sensor"),
                      0x01: _("Motion Sensor"),
                      0x02: _("Air quality Sensor"),
                      0x03: _("Temperature Sensor"),
                      0x04: _("Humidity Sensor"),
                      0x05: _("Leak Sensor"),
                      0x06: _("Smoke Sensor"),
                      0x07: _("Occupancy Sensor"),
                      0x08: _("Contact Sensor"),
                      0x09: _("Carbon Monoxide Sensor"),
                      0x0A: _("Carbon Dioxide Sensor"),
                      0x0B: _("Ambient Light Sensor"),
                      0x0C: _("Energy Sensor"),
                      0x0D: _("Color Light Sensor"),
                      0x0E: _("Rain Sensor"),
                      0x0F: _("Fire Sensor"),
                      0x10: _("Wind Sensor"),
                      0x11: _("Proximity Sensor"),
                      0x12: _("Multi-Sensor"),
                      0x13: _("Flush Mounted Sensor"),
                      0x14: _("Ceiling Mounted Sensor"),
                      0x15: _("Wall Mounted Sensor"),
                      0x16: _("Multisensor"),
                      0x17: _("Energy Meter"),
                      0x18: _("Flame Detector"),
                      0x19: _("Vehicle Tire Pressure Sensor")}),
    0x16: ("Light Fixtures", {0x00: _("Generic Light Fixtures"),
                              0x01: _("Wall Light"),
                              0x02: _("Ceiling Light"),
                              0x03: _("Floor Light"),
                              0x04: _("Cabinet Light"),
                              0x05: _("Desk Light"),
                              0x06: _("Troffer Light"),
                              0x07: _("Pendant Light"),
                              0x08: _("In-ground Light"),
                              0x09: _("Flood Light"),
                              0x0A: _("Underwater Light"),
                              0x0B: _("Bollard with Light"),
                              0x0C: _("Pathway Light"),
                              0x0D: _("Garden Light"),
                              0x0E: _("Pole-top Light"),
                              0x0F: _("Spotlight"),
                              0x10: _("Linear Light"),
                              0x11: _("Street Light"),
                              0x12: _("Shelves Light"),
                              0x13: _("Bay Light"),
                              0x14: _("Emergency Exit Light"),
                              0x15: _("Light Controller"),
                              0x16: _("Light Driver"),
                              0x17: _("Bulb"),
                              0x18: _("Low-bay Light"),
                              0x19: _("High-bay Light")}),
    0x17: ("Fan", {0x00: _("Generic Fan"),
                   0x01: _("Ceiling Fan"),
                   0x02: _("Axial Fan"),
                   0x03: _("Exhaust Fan"),
                   0x04: _("Pedestal Fan"),
                   0x05: _("Desk Fan"),
                   0x06: _("Wall Fan")}),
    0x18: ("HVAC", {0x00: _("Generic HVAC"),
                    0x01: _("Thermostat"),
                    0x02: _("Humidifier"),
                    0x03: _("De-humidifier"),
                    0x04: _("Heater"),
                    0x05: _("Radiator"),
                    0x06: _("Boiler"),
                    0x07: _("Heat Pump"),
                    0x08: _("Infrared Heater"),
                    0x09: _("Radiant Panel Heater"),
                    0x0A: _("Fan Heater"),
                    0x0B: _("Air Curtain")}),
    0x19: ("Air Conditioning", {0x00: _("Generic Air Conditioning")}),
    0x1A: ("Humidifier", {0x00: _("Generic Humidifier")}),
    0x1B: ("Heating", {0x00: _("Generic Heating"),
                       0x01: _("Radiator"),
                       0x02: _("Boiler"),
                       0x03: _("Heat Pump"),
                       0x04: _("Infrared Heater"),
                       0x05: _("Radiant Panel Heater"),
                       0x06: _("Fan Heater"),
                       0x07: _("Air Curtain")}),
    0x1C: ("Access Control", {0x00: _("Generic Access Control"),
                              0x01: _("Access Door"),
                              0x02: _("Garage Door"),
                              0x03: _("Emergency Exit Door"),
                              0x04: _("Access Lock"),
                              0x05: _("Elevator"),
                              0x06: _("Window"),
                              0x07: _("Entrance Gate"),
                              0x08: _("Door Lock"),
                              0x09: _("Locker")}),
    0x1D: ("Motorized Device", {0x00: _("Generic Motorized Device"),
                                0x01: _("Motorized Gate"),
                                0x02: _("Awning"),
                                0x03: _("Blinds or Shades"),
                                0x04: _("Curtains"),
                                0x05: _("Screen")}),
    0x1E: ("Power Device", {0x00: _("Generic Power Device"),
                            0x01: _("Power Outlet"),
                            0x02: _("Power Strip"),
                            0x03: _("Plug"),
                            0x04: _("Power Supply"),
                            0x05: _("LED Driver"),
                            0x06: _("Fluorescent Lamp Gear"),
                            0x07: _("HID Lamp Gear"),
                            0x08: _("Charge Case"),
                            0x09: _("Power Bank")}),
    0x1F: ("Light Source", {0x00: _("Generic Light Source"),
                            0x01: _("Incandescent Light Bulb"),
                            0x02: _("LED Lamp"),
                            0x03: _("HID Lamp"),
                            0x04: _("Fluorescent Lamp"),
                            0x05: _("LED Array"),
                            0x06: _("Multi-Color LED Array"),
                            0x07: _("Low voltage halogen"),
                            0x08: _("Organic light emitting diode (OLED)")}),
    0x20: ("Window Covering", {0x00: _("Generic Window Covering"),
                               0x01: _("Window Shades"),
                               0x02: _("Window Blinds"),
                               0x03: _("Window Awning"),
                               0x04: _("Window Curtain"),
                               0x05: _("Exterior Shutter"),
                               0x06: _("Exterior Screen")}),
    0x21: ("Audio Sink", {0x00: _("Generic Audio Sink"),
                          0x01: _("Standalone Speaker"),
                          0x02: _("Soundbar"),
                          0x03: _("Bookshelf Speaker"),
                          0x04: _("Standmounted Speaker"),
                          0x05: _("Speakerphone")}),
    0x22: ("Audio Source", {0x00: _("Generic Audio Source"),
                            0x01: _("Microphone"),
                            0x02: _("Alarm"),
                            0x03: _("Bell"),
                            0x04: _("Horn"),
                            0x05: _("Broadcasting Device"),
                            0x06: _("Service Desk"),
                            0x07: _("Kiosk"),
                            0x08: _("Broadcasting Room"),
                            0x09: _("Auditorium")}),
    0x23: ("Motorized Vehicle", {0x00: _("Generic Motorized Vehicle"),
                                 0x01: _("Car"),
                                 0x02: _("Large Goods Vehicle"),
                                 0x03: _("2-Wheeled Vehicle"),
                                 0x04: _("Motorbike"),
                                 0x05: _("Scooter"),
                                 0x06: _("Moped"),
                                 0x07: _("3-Wheeled Vehicle"),
                                 0x08: _("Light Vehicle"),
                                 0x09: _("Quad Bike"),
                                 0x0A: _("Minibus"),
                                 0x0B: _("Bus"),
                                 0x0C: _("Trolley"),
                                 0x0D: _("Agricultural Vehicle"),
                                 0x0E: _("Camper / Caravan"),
                                 0x0F: _("Recreational Vehicle / Motor Home")}),
    0x24: ("Domestic Appliance", {0x00: _("Generic Domestic Appliance"),
                                  0x01: _("Refrigerator"),
                                  0x02: _("Freezer"),
                                  0x03: _("Oven"),
                                  0x04: _("Microwave"),
                                  0x05: _("Toaster"),
                                  0x06: _("Washing Machine"),
                                  0x07: _("Dryer"),
                                  0x08: _("Coffee maker"),
                                  0x09: _("Clothes iron"),
                                  0x0A: _("Curling iron"),
                                  0x0B: _("Hair dryer"),
                                  0x0C: _("Vacuum cleaner"),
                                  0x0D: _("Robotic vacuum cleaner"),
                                  0x0E: _("Rice cooker"),
                                  0x0F: _("Clothes steamer")}),
    0x25: ("Wearable Audio Device", {0x00: _("Generic Wearable Audio Device"),
                                     0x01: _("Earbud"),
                                     0x02: _("Headset"),
                                     0x03: _("Headphones"),
                                     0x04: _("Neck Band")}),
    0x26: ("Aircraft", {0x00: _("Generic Aircraft"),
                        0x01: _("Light Aircraft"),
                        0x02: _("Microlight"),
                        0x03: _("Paraglider"),
                        0x04: _("Large Passenger Aircraft")}),
    0x27: ("AV Equipment", {0x00: _("Generic AV Equipment"),
                            0x01: _("Amplifier"),
                            0x02: _("Receiver"),
                            0x03: _("Radio"),
                            0x04: _("Tuner"),
                            0x05: _("Turntable"),
                            0x06: _("CD Player"),
                            0x07: _("DVD Player"),
                            0x08: _("Bluray Player"),
                            0x09: _("Optical Disc Player"),
                            0x0A: _("Set-Top Box")}),
    0x28: ("Display Equipment", {0x00: _("Generic Display Equipment"),
                                 0x01: _("Television"),
                                 0x02: _("Monitor"),
                                 0x03: _("Projector")}),
    0x29: ("Hearing aid", {0x00: _("Generic Hearing aid"),
                           0x01: _("In-ear hearing aid"),
                           0x02: _("Behind-ear hearing aid"),
                           0x03: _("Cochlear Implant")}),
    0x2A: ("Gaming", {0x00: _("Generic Gaming"),
                      0x01: _("Home Video Game Console"),
                      0x02: _("Portable handheld console")}),
    0x2B: ("Signage", {0x00: _("Generic Signage"),
                       0x01: _("Digital Signage"),
                       0x02: _("Electronic Label")}),
    0x31: ("Pulse Oximeter", {0x00: _("Generic Pulse Oximeter"),
                              0x01: _("Fingertip Pulse Oximeter"),
                              0x02: _("Wrist Worn Pulse Oximeter")}),
    0x32: ("Weight Scale", {0x00: _("Generic Weight Scale")}),
    0x33: ("Personal Mobility Device", {0x00: _("Generic Personal Mobility Device"),
                                        0x01: _("Powered Wheelchair"),
                                        0x02: _("Mobility Scooter")}),
    0x34: ("Continuous Glucose Monitor", {0x00: _("Generic Continuous Glucose Monitor")}),
    0x35: ("Insulin Pump", {0x00: _("Generic Insulin Pump"),
                            0x01: _("Insulin Pump, durable pump"),
                            0x04: _("Insulin Pump, patch pump"),
                            0x08: _("Insulin Pen")}),
    0x36: ("Medication Delivery", {0x00: _("Generic Medication Delivery")}),
    0x37: ("Spirometer", {0x00: _("Generic Spirometer"),
                          0x01: _("Handheld Spirometer")}),
    0x51: ("Outdoor Sports Activity", {0x00: _("Generic Outdoor Sports Activity"),
                                       0x01: _("Location Display"),
                                       0x02: _("Location and Navigation Display"),
                                       0x03: _("Location Pod"),
                                       0x04: _("Location and Navigation Pod")})
}


def get_major_class(klass: int) -> str:
    index = (klass >> 8) & 0x1F

    if index > 8:
        return major_cls[9]

    return major_cls[index]


def get_minor_class(klass: int) -> str:
    if not klass:
        return _("Unknown")

    i = (klass >> 8) & 0x1F

    if i == 1:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(computer_minor_cls):
            return computer_minor_cls[minor_index]
        else:
            return ""
    elif i == 2:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(phone_minor_cls):
            return phone_minor_cls[minor_index]
        return ""
    elif i == 3:
        minor_index = (klass >> 5) & 0x07
        if minor_index < len(access_point_minor_cls):
            return access_point_minor_cls[minor_index]
        else:
            return ""
    elif i == 4:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(audio_video_minor_cls):
            return audio_video_minor_cls[minor_index]
        else:
            return ""
    elif i == 5:
        minor_index = (klass >> 6) & 0x03
        if minor_index < len(peripheral_minor_cls):
            return peripheral_minor_cls[minor_index]
        else:
            return ""
    elif i == 6:
        return _("Imaging")

    elif i == 7:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(wearable_minor_cls):
            return wearable_minor_cls[minor_index]
        else:
            return ""
    elif i == 8:
        minor_index = (klass >> 2) & 0x3F
        if minor_index < len(toy_minor_cls):
            return toy_minor_cls[minor_index]
        else:
            return ""

    return _("Unknown")


# First 10 bits is the category, the following 6 bits sub category
def gatt_appearance_to_name(appearance: int) -> str:
    cat = appearance >> 0x6
    subcat = appearance & 0x3f

    if (19 <= cat <= 48) or (55 <= cat <= 80):
        # These ranges are reserved
        logging.debug(f"Reserved category found: {appearance}")
        return gatt_appearance_categories[0][0]
    elif cat > 81:
        logging.warning(f"Invalid category found: {appearance}")
        return gatt_appearance_categories[0][0]

    cat_name, subcats = gatt_appearance_categories[cat]

    if subcat not in subcats:
        return cat_name
    else:
        return subcats[subcat]



import os.path
import shutil
import subprocess
from ipaddress import IPv4Address
from subprocess import Popen
from typing import Optional, List
from unittest import TestCase
from unittest.mock import patch, Mock, PropertyMock

from blueman.main.NetConf import DnsMasqHandler, NetworkSetupError, DhcpdHandler, UdhcpdHandler, NetConf, DHCPHandler


class FakeSocket:
    def __init__(self, connect_return_value: int) -> None:
        self._ret = connect_return_value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def connect_ex(self, *_args: object) -> int:
        return self._ret


@patch("blueman.main.NetConf.have", return_value="/usr/bin/mydnsmasq")
@patch("blueman.main.NetConf.DnsMasqHandler._pid_path", PropertyMock(return_value="/tmp/pid"))
class TestDnsmasqHandler(TestCase):
    @patch("blueman.main.NetConf.Popen", return_value=Popen("true"))
    @patch("blueman.main.NetConf.NetConf.lock")
    @patch("blueman.main.NetConf.socket.socket", lambda *args: FakeSocket(1))
    @patch("blueman.main.NetConf.DNSServerProvider.get_servers", lambda: [])
    def test_success_with_dns(self, lock_mock: Mock, popen_mock: Mock, have_mock: Mock) -> None:
        with open("/tmp/pid", "w") as f:
            f.write("123")
        DnsMasqHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        lock_mock.assert_called_with("dhcp")

    @patch("blueman.main.NetConf.Popen", return_value=Popen("true"))
    @patch("blueman.main.NetConf.NetConf.lock")
    @patch("blueman.main.NetConf.socket.socket", lambda *args: FakeSocket(0))
    @patch("blueman.main.NetConf.DNSServerProvider.get_servers", lambda: [IPv4Address("203.0.113.10")])
    def test_success_without_dns(self, lock_mock: Mock, popen_mock: Mock, have_mock: Mock) -> None:
        with open("/tmp/pid", "w") as f:
            f.write("123")
        DnsMasqHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock, ["--port=0", "--dhcp-option=option:dns-server,203.0.113.10"])
        lock_mock.assert_called_with("dhcp")

    @patch("blueman.main.NetConf.Popen", return_value=Popen(["sh", "-c", "echo errormsg >&2"], stderr=subprocess.PIPE))
    @patch("blueman.main.NetConf.socket.socket", lambda *args: FakeSocket(1))
    @patch("blueman.main.NetConf.DNSServerProvider.get_servers", lambda: [])
    def test_failure(self, popen_mock: Mock, have_mock: Mock) -> None:
        with self.assertRaises(NetworkSetupError) as cm:
            DnsMasqHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        self.assertEqual(cm.exception.args, ("dnsmasq failed to start: errormsg",))

    def _check_invocation(self, have_mock: Mock, popen_mock: Mock, additional_args: Optional[List[str]] = None) -> None:
        have_mock.assert_called_with("dnsmasq")
        popen_mock.assert_called_with(
            ["/usr/bin/mydnsmasq", "--pid-file=/tmp/pid", "--except-interface=lo",
             "--interface=pan1", "--bind-interfaces", "--dhcp-range=203.0.113.2,203.0.113.254,60m",
             "--dhcp-option=option:router,203.0.113.1"] + ([] if additional_args is None else additional_args),
            stderr=subprocess.PIPE
        )


@patch("blueman.main.NetConf.have", return_value="/usr/bin/mydhcpd")
@patch("blueman.main.NetConf.DhcpdHandler._pid_path", PropertyMock(return_value="/tmp/pid"))
@patch("blueman.main.NetConf.DHCP_CONFIG_FILE", "/tmp/dhcpd.conf")
@patch("blueman.main.NetConf.DNSServerProvider.get_servers", lambda: [])
class TestDhcpdHandler(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open("/tmp/dhcpd.conf", "w"):
            pass

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove("/tmp/dhcpd.conf")

    @patch("blueman.main.NetConf.Popen", return_value=Popen(["sh", "-c", "echo warning >&2"], stderr=subprocess.PIPE))
    @patch("blueman.main.NetConf.NetConf.lock")
    def test_success(self, lock_mock: Mock, popen_mock: Mock, have_mock: Mock) -> None:
        with open("/tmp/pid", "w") as f:
            f.write("123")
        DhcpdHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        lock_mock.assert_called_with("dhcp")

    @patch("blueman.main.NetConf.Popen",
           return_value=Popen(["sh", "-c", "echo errormsg >&2; exit 1"], stderr=subprocess.PIPE))
    def test_failure(self, popen_mock: Mock, have_mock: Mock) -> None:
        with self.assertRaises(NetworkSetupError) as cm:
            DhcpdHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        self.assertEqual(cm.exception.args, ("dhcpd failed to start: errormsg",))

    def _check_invocation(self, have_mock: Mock, popen_mock: Mock) -> None:
        have_mock.assert_called_with("dhcpd3")
        popen_mock.assert_called_with(
            ["/usr/bin/mydhcpd", "-pf", "/tmp/pid", "pan1"],
            stderr=subprocess.PIPE
        )


@patch("blueman.main.NetConf.have", return_value="/usr/bin/myudhcpd")
@patch("blueman.main.NetConf.UdhcpdHandler._pid_path", PropertyMock(return_value="/tmp/pid"))
@patch("blueman.main.NetConf.DNSServerProvider.get_servers", lambda: [])
class TestUdhcpdHandler(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open("/tmp/dhcpd.conf", "w"):
            pass

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove("/tmp/dhcpd.conf")

    @patch("blueman.main.NetConf.Popen", return_value=Popen(["sh", "-c", "echo warning >&2"], stderr=subprocess.PIPE))
    @patch("blueman.main.NetConf.NetConf.lock")
    @patch("blueman.main.NetConf._is_running", lambda _name, _pid: True)
    def test_success(self, lock_mock: Mock, popen_mock: Mock, have_mock: Mock) -> None:
        with open("/tmp/pid", "w") as f:
            f.write("123")
        UdhcpdHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        lock_mock.assert_called_with("dhcp")

    @patch("blueman.main.NetConf.Popen", return_value=Popen(["sh", "-c", "echo errormsg >&2"], stderr=subprocess.PIPE))
    @patch("blueman.main.NetConf._is_running", lambda _name, _pid: False)
    def test_failure(self, popen_mock: Mock, have_mock: Mock) -> None:
        with open("/tmp/pid", "w") as f:
            f.write("123")
        with self.assertRaises(NetworkSetupError) as cm:
            UdhcpdHandler().apply("203.0.113.1", "255.255.255.0")
        self._check_invocation(have_mock, popen_mock)
        self.assertEqual(cm.exception.args, ("udhcpd failed to start: errormsg",))

    def _check_invocation(self, have_mock: Mock, popen_mock: Mock) -> None:
        have_mock.assert_called_with("udhcpd")
        args = popen_mock.call_args
        self.assertEqual(len(args[0]), 1)
        self.assertEqual(args[0][0][:-1], ["/usr/bin/myudhcpd", "-S"])
        self.assertEqual(args[1], {"stderr": subprocess.PIPE})


@patch("blueman.main.NetConf.NetConf._IPV4_SYS_PATH", "/tmp/blueman-test/ipv4")
@patch("blueman.main.NetConf.NetConf._RUN_PATH", "/tmp/blueman-test/run")
@patch("blueman.main.NetConf.create_bridge")
@patch("blueman.main.NetConf.call", return_value=0)
class TestNetConf(TestCase):
    def setUp(self) -> None:
        os.makedirs("/tmp/blueman-test/run", exist_ok=True)
        os.makedirs("/tmp/blueman-test/ipv4/conf", exist_ok=True)
        os.makedirs("/tmp/blueman-test/ipv4/conf/i0", exist_ok=True)
        os.makedirs("/tmp/blueman-test/ipv4/conf/i1", exist_ok=True)

    def tearDown(self) -> None:
        shutil.rmtree("/tmp/blueman-test")

    class TestDHCPHandler(DHCPHandler):
        @property
        def _key(self) -> str:
            return "test-dhcp-handler"

        apply = Mock()

    class TestDHCPHandler2(TestDHCPHandler):
        clean_up = Mock()

    def test_initial_apply(self, call_mock: Mock, bridge_mock: Mock) -> None:
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler, False)

        bridge_mock.assert_called_once_with("pan1")
        self._check_forwarding()
        self._check_iptables(call_mock)
        self.TestDHCPHandler.apply.assert_called_with("203.0.113.1", "255.255.255.0")

        call_mock.assert_any_call(["ip", "link", "set", "dev", "pan1", "up"])
        call_mock.assert_any_call(["ip", "address", "add", "203.0.113.1/255.255.255.0", "dev", "pan1"])
        self.assertTrue(NetConf.locked("netconfig"))

    def test_address_change(self, call_mock: Mock, bridge_mock: Mock) -> None:
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler, True)

        bridge_mock.assert_called_once_with("pan1")
        self._check_forwarding()
        self._check_iptables(call_mock)
        self.TestDHCPHandler.apply.assert_called_with("203.0.113.1", "255.255.255.0")

        call_mock.assert_any_call(["ip", "link", "set", "dev", "pan1", "up"])
        call_mock.assert_any_call(["ip", "address", "add", "203.0.113.1/255.255.255.0", "dev", "pan1"])
        self.assertTrue(NetConf.locked("netconfig"))

    @patch("blueman.main.NetConf.have", lambda key: key != "ip")
    def test_nettools(self, call_mock: Mock, bridge_mock: Mock) -> None:
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler, False)

        bridge_mock.assert_called_once_with("pan1")
        self._check_forwarding()

        call_mock.assert_any_call(["ifconfig", "pan1", "203.0.113.1", "netmask", "255.255.255.0", "up"])
        self.assertTrue(NetConf.locked("netconfig"))

    def _check_forwarding(self) -> None:
        for file in [
            "/tmp/blueman-test/ipv4/ip_forward",
            "/tmp/blueman-test/ipv4/conf/i0/forwarding",
            "/tmp/blueman-test/ipv4/conf/i1/forwarding",
        ]:
            with open(file, "r") as f:
                self.assertEqual(f.read(), "1")

    def _check_iptables(self, call_mock: Mock, remove: bool = False) -> None:
        command = "-D" if remove else "-A"
        call_mock.assert_any_call(["/sbin/iptables", "-t", "nat", command, "POSTROUTING",
                                   "-s", "203.0.113.1/255.255.255.0", "-j", "MASQUERADE"])
        call_mock.assert_any_call(["/sbin/iptables", "-t", "filter", command, "FORWARD", "-i", "pan1", "-j", "ACCEPT"])
        call_mock.assert_any_call(["/sbin/iptables", "-t", "filter", command, "FORWARD", "-o", "pan1", "-j", "ACCEPT"])
        self.assertEqual(NetConf.locked("iptables"), not remove)

    def test_dhcp_handler_replacement(self, _call_mock: Mock, _bridge_mock: Mock) -> None:
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler2, False)
        self.TestDHCPHandler2.clean_up.reset_mock()
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler, False)
        self.TestDHCPHandler2.clean_up.assert_called_once_with()
        self.TestDHCPHandler.apply.assert_called_with("203.0.113.1", "255.255.255.0")

    @patch("blueman.main.NetConf.destroy_bridge")
    def test_cleanup(self, destroy_bridge_mock: Mock, call_mock: Mock, _create_bridge_mock: Mock) -> None:
        NetConf.apply_settings("203.0.113.1", "255.255.255.0", self.TestDHCPHandler2, False)
        self.TestDHCPHandler2.clean_up.reset_mock()

        NetConf.clean_up()

        destroy_bridge_mock.assert_called_once_with("pan1")
        self.TestDHCPHandler2.clean_up.assert_called_once_with()
        self._check_iptables(call_mock, remove=True)

        self.assertFalse(NetConf.locked("netconfig"))
        self.assertFalse(NetConf.locked("iptables"))



from gettext import gettext as _
import logging

from gi.repository import Gio, GLib

from blueman.gobject import SingletonGObjectMeta
from blueman.gui.Notification import Notification


class DBusProxyFailed(Exception):
    pass


class ProxyBase(Gio.DBusProxy, metaclass=SingletonGObjectMeta):
    def __init__(self, name: str, interface_name: str, object_path: str = "/", systembus: bool = False,
                 flags: Gio.DBusProxyFlags = Gio.DBusProxyFlags.NONE) -> None:
        if systembus:
            bustype = Gio.BusType.SYSTEM
        else:
            bustype = Gio.BusType.SESSION

        super().__init__(
            g_name=name,
            g_interface_name=interface_name,
            g_object_path=object_path,
            g_bus_type=bustype,
            g_flags=flags
        )

        try:
            self.init()
        except GLib.Error as e:
            raise DBusProxyFailed(e.message)

    def call_method(self, name: str, params: GLib.Variant) -> None:
        def call_finish(proxy: ProxyBase, response: Gio.AsyncResult) -> None:
            try:
                proxy.call_finish(response)
            except GLib.Error:
                logging.error(f"Failed to execute method {name}", exc_info=True)
                raise

        self.call(name, params, Gio.DBusCallFlags.NONE, -1, None, call_finish)


class Mechanism(ProxyBase):
    def __init__(self) -> None:
        super().__init__(name='org.blueman.Mechanism', interface_name='org.blueman.Mechanism',
                         object_path="/org/blueman/mechanism", systembus=True)


class AppletService(ProxyBase):
    def __init__(self) -> None:
        super().__init__(name='org.blueman.Applet', interface_name='org.blueman.Applet',
                         object_path="/org/blueman/Applet")


class ManagerService(ProxyBase):
    def __init__(self) -> None:
        super().__init__(name="org.blueman.Manager", interface_name="org.freedesktop.Application",
                         object_path="/org/blueman/Manager",
                         flags=Gio.DBusProxyFlags.DO_NOT_AUTO_START_AT_CONSTRUCTION)

    def activate(self) -> None:
        try:
            param = GLib.Variant('(a{sv})', ({},))
            self.call_method("Activate", param)
        except GLib.Error:
            # This can different errors, depending on the system configuration, typically:
            # * org.freedesktop.DBus.Error.Spawn.ChildExited if dbus-daemon tries to launch the service itself.
            # * org.freedesktop.DBus.Error.NoReply if the systemd integration is used.
            Notification(
                _("Failed to reach blueman-manager"),
                _("It seems like blueman-manager could no get activated via D-Bus. "
                  "A typical cause for this is a broken graphical setup in the D-Bus activation environment "
                  "that can get resolved with a call to dbus-update-activation-environment, "
                  "typically issued from xinitrc (respectively the Sway config or similar)."),
                0,
            ).show()

    def _call_action(self, name: str) -> None:
        param = GLib.Variant('(sava{sv})', (name, [], {}))
        self.call_method('ActivateAction', param)

    def quit(self) -> None:
        self._call_action("Quit")



from blueman.bluez.AnyBase import AnyBase

from blueman.bluez.Base import Base

_INTERFACE = "org.bluez.Battery1"


class Battery(Base):
    _interface_name = _INTERFACE


class AnyBattery(AnyBase):
    def __init__(self) -> None:
        super().__init__(_INTERFACE)



import weakref
from typing import Callable

from blueman.bluez.Battery import Battery, AnyBattery
from blueman.bluez.Manager import Manager


class BatteryWatcher:
    def __init__(self, callback: Callable[[str, int], None]) -> None:
        super().__init__()
        manager = Manager()
        weakref.finalize(
            self,
            manager.disconnect_signal,
            manager.connect_signal(
                "battery-created",
                lambda _manager, obj_path: callback(obj_path, Battery(obj_path=obj_path)["Percentage"])
            )
        )

        any_battery = AnyBattery()
        weakref.finalize(
            self,
            any_battery.disconnect_signal,
            any_battery.connect_signal(
                "property-changed",
                lambda _any_battery, key, value, path: callback(path, value) if key == "Percentage" else None
            )
        )



import os
import logging
import importlib
import signal
from typing import Optional

from blueman.main.Builder import Builder
import blueman.plugins.services
from blueman.plugins.ServicePlugin import ServicePlugin

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import Gio


class BluemanServices(Gtk.Application):
    def __init__(self) -> None:
        super().__init__(application_id="org.blueman.Services")
        self.window: Optional[Gtk.Window] = None
        self.builder = Builder("services-window.ui")
        self.Config = Gio.Settings(schema_id="org.blueman.general")

        def do_quit(_: object) -> bool:
            self.quit()
            return False

        s = GLib.unix_signal_source_new(signal.SIGINT)
        s.set_callback(do_quit)
        s.attach()

        close_action = Gio.SimpleAction.new("quit", None)
        close_action.connect("activate", lambda x, y: self.quit())
        self.add_action(close_action)

    def do_activate(self) -> None:
        if not self.window:
            self.window = self.builder.get_widget("window", Gtk.ApplicationWindow)
            self.window.set_application(self)
            self.notebook = self.builder.get_widget("notebook", Gtk.Notebook)
            self.b_apply = self.builder.get_widget("apply", Gtk.Button)
            self.b_apply.connect("clicked", self.on_apply_clicked)

            self.load_plugins()

            self.Config.bind("services-last-item", self.notebook, "page", Gio.SettingsBindFlags.DEFAULT)

        self.window.present_with_time(Gtk.get_current_event_time())

    def do_startup(self) -> None:
        Gtk.Application.do_startup(self)
        self.set_accels_for_action("app.quit", ["<Ctrl>w", "<Ctrl>q", "Escape"])

    def option_changed(self) -> None:
        rets = [plugin.on_query_apply_state() for plugin in ServicePlugin.instances]
        show_apply = False
        for ret in rets:
            if ret == -1:
                show_apply = False
                break
            assert isinstance(ret, bool)
            show_apply = show_apply or ret

        self.b_apply.props.sensitive = show_apply

    def load_plugins(self) -> None:
        path = os.path.dirname(blueman.plugins.services.__file__)
        plugins = []
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith(".py") and not (f.endswith(".pyc") or f.endswith("_.py")):
                    plugins.append(f[0:-3])
        plugins.sort()
        logging.info(plugins)
        for plugin in plugins:
            try:
                importlib.import_module(f"blueman.plugins.services.{plugin}")
            except ImportError:
                logging.error(f"Unable to load {plugin} plugin", exc_info=True)

        for cls in ServicePlugin.__subclasses__():
            # FIXME this should not fail, if it does its a bug in the plugin
            try:
                inst = cls(self)
            except:  # noqa: E722
                logging.error(f"Failed to create instance of {cls}", exc_info=True)
                continue
            if not cls.__plugin_info__:
                logging.warning(f"Invalid plugin info in {cls}")
            else:
                (name, icon) = cls.__plugin_info__
                inst.on_load()
                self.add_page(inst, name, icon)

    def add_page(self, inst: ServicePlugin, name: str, icon_name: str) -> None:
        box = Gtk.Box(vexpand=True, visible=True, spacing=6)
        label = Gtk.Label(label=name, visible=True)
        icon = Gtk.Image(icon_name=icon_name, visible=True)
        box.add(icon)
        box.add(label)
        self.notebook.append_page(inst.widget, box)

    def on_apply_clicked(self, _button: Gtk.Button) -> None:
        for plugin in ServicePlugin.instances:
            plugin.on_apply()
        self.option_changed()



import gi
gi.require_version("Gtk", "3.0")

from gi.repository import Gio, GLib, Gtk
import logging
import signal
from typing import Any, cast

from blueman.Functions import *
from blueman.bluez.Manager import Manager
from blueman.bluez.Adapter import AnyAdapter
from blueman.bluez.Device import AnyDevice
import blueman.plugins.applet
from blueman.main.PluginManager import PersistentPluginManager
from blueman.main.DbusService import DbusService
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.applet.DBusService import DBusService
from blueman.plugins.applet.Menu import Menu
from blueman.plugins.applet.PowerManager import PowerManager
from blueman.plugins.applet.RecentConns import RecentConns
from blueman.plugins.applet.StandardItems import StandardItems
from blueman.plugins.applet.StatusIcon import StatusIcon


class BluemanApplet(Gtk.Application):
    def __init__(self) -> None:
        super().__init__(application_id="org.blueman.Applet", flags=Gio.ApplicationFlags.FLAGS_NONE)
        setup_icon_path()

        def do_quit(_: object) -> bool:
            self.quit()
            return False

        s = GLib.unix_signal_source_new(signal.SIGINT)
        s.set_callback(do_quit)
        s.attach()

        self.plugin_run_state_changed = False
        self.manager_state = False
        self._active = False

        self.Manager = Manager()
        self.Manager.connect_signal('adapter-added', self.on_adapter_added)
        self.Manager.connect_signal('adapter-removed', self.on_adapter_removed)
        self.Manager.connect_signal('device-created', self.on_device_created)
        self.Manager.connect_signal('device-removed', self.on_device_removed)

        self.DbusSvc = DbusService("org.blueman.Applet", "org.blueman.Applet", "/org/blueman/Applet",
                                   Gio.BusType.SESSION)
        self.DbusSvc.register()

        self.Plugins = Plugins(self)
        self.Plugins.load_plugin()

        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_plugins_loaded()

        self.Manager.watch_name_owner(self._on_dbus_name_appeared, self._on_dbus_name_vanished)

        self._any_adapter = AnyAdapter()
        self._any_adapter.connect_signal('property-changed', self._on_adapter_property_changed)

        self._any_device = AnyDevice()
        self._any_device.connect_signal('property-changed', self._on_device_property_changed)

    def do_startup(self) -> None:
        Gtk.Application.do_startup(self)
        self.set_accels_for_action("win.close", ["<Ctrl>w", "Escape"])

    def do_activate(self) -> None:
        if not self._active:
            self.hold()
            self._active = True

    def _on_dbus_name_appeared(self, _connection: Gio.DBusConnection, name: str, owner: str) -> None:
        logging.info(f"{name} {owner}")
        self.manager_state = True
        self.plugin_run_state_changed = True
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_manager_state_changed(self.manager_state)

    def _on_dbus_name_vanished(self, _connection: Gio.DBusConnection, name: str) -> None:
        logging.info(name)
        self.manager_state = False
        self.plugin_run_state_changed = True
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_manager_state_changed(self.manager_state)

    def _on_adapter_property_changed(self, _adapter: AnyAdapter, key: str, value: Any, path: str) -> None:
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_adapter_property_changed(path, key, value)

    def _on_device_property_changed(self, _device: AnyDevice, key: str, value: Any, path: str) -> None:
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_device_property_changed(path, key, value)

    def on_adapter_added(self, _manager: Manager, path: str) -> None:
        logging.info(f"Adapter added {path}")
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_adapter_added(path)

    def on_adapter_removed(self, _manager: Manager, path: str) -> None:
        logging.info(f"Adapter removed {path}")
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_adapter_removed(path)

    def on_device_created(self, _manager: Manager, path: str) -> None:
        logging.info(f"Device created {path}")
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_device_created(path)

    def on_device_removed(self, _manager: Manager, path: str) -> None:
        logging.info(f"Device removed {path}")
        for plugin in self.Plugins.get_loaded_plugins(AppletPlugin):
            plugin.on_device_removed(path)


class Plugins(PersistentPluginManager[AppletPlugin]):
    def __init__(self, applet: BluemanApplet):
        super().__init__(AppletPlugin, blueman.plugins.applet, applet)

    @property
    def DBusService(self) -> DBusService:
        return cast(DBusService, self._plugins["DBusService"])

    @property
    def Menu(self) -> Menu:
        return cast(Menu, self._plugins["Menu"])

    @property
    def PowerManager(self) -> PowerManager:
        return cast(PowerManager, self._plugins["PowerManager"])

    @property
    def RecentConns(self) -> RecentConns:
        return cast(RecentConns, self._plugins["RecentConns"])

    @property
    def StandardItems(self) -> StandardItems:
        return cast(StandardItems, self._plugins["StandardItems"])

    @property
    def StatusIcon(self) -> StatusIcon:
        return cast(StatusIcon, self._plugins["StatusIcon"])



import logging
from gettext import gettext as _
from typing import Any, Dict, Union

from blueman.bluez.Device import Device
from blueman.gui.Notification import Notification, _NotificationBubble, _NotificationDialog
from blueman.main.BatteryWatcher import BatteryWatcher
from blueman.plugins.AppletPlugin import AppletPlugin
from gi.repository import GLib


class ConnectionNotifier(AppletPlugin):
    __author__ = "cschramm"
    __icon__ = "bluetooth-symbolic"
    __description__ = _("Shows desktop notifications when devices get connected or disconnected.")

    _notifications: Dict[str, Union[_NotificationBubble, _NotificationDialog]] = {}

    def on_load(self) -> None:
        self._battery_watcher = BatteryWatcher(self._on_battery_update)

    def on_unload(self) -> None:
        del self._battery_watcher

    def on_device_property_changed(self, path: str, key: str, value: Any) -> None:
        if key == "Connected":
            device = Device(obj_path=path)
            if value:
                self._notifications[path] = notification = Notification(
                    device.display_name,
                    _('Connected'),
                    icon_name=device["Icon"]
                )
                notification.show()
            else:
                Notification(device.display_name, _('Disconnected'), icon_name=device["Icon"]).show()

    def _on_battery_update(self, path: str, value: int) -> None:
        notification = self._notifications.pop(path, None)
        if notification:
            try:
                notification.set_message(f"{_('Connected')} {value}%")
                notification.set_notification_icon("battery")
            except GLib.Error:
                logging.error("Failed to update notification", exc_info=True)



import re
import socket
from ipaddress import IPv4Address, ip_address
from typing import List, Generator

from gi.repository import GObject, Gio, GLib

from blueman.bluemantyping import GSignals


class DNSServerProvider(GObject.GObject):
    RESOLVER_PATH = "/etc/resolv.conf"
    _RESOLVED_NAME = "org.freedesktop.resolve1"
    _RESOLVED_MANAGER_INTERFACE = "org.freedesktop.resolve1.Manager"

    __gsignals__: GSignals = {"changed": (GObject.SignalFlags.NO_HOOKS, None, ())}

    def __init__(self) -> None:
        super().__init__()
        self._subscribe_systemd_resolved()
        self._subscribe_resolver()

    @classmethod
    def get_servers(cls) -> List[IPv4Address]:
        return list(set(cls._get_servers_from_systemd_resolved()) or cls._get_servers_from_resolver())

    @classmethod
    def _get_servers_from_systemd_resolved(cls) -> Generator[IPv4Address, None, None]:
        bus = Gio.bus_get_sync(Gio.BusType.SYSTEM)

        manager_proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SYSTEM,
            Gio.DBusProxyFlags.DO_NOT_AUTO_START,
            None,
            cls._RESOLVED_NAME,
            "/org/freedesktop/resolve1",
            cls._RESOLVED_MANAGER_INTERFACE,
        )

        try:
            data = manager_proxy.call_sync(
                "org.freedesktop.DBus.Properties.Get",
                GLib.Variant("(ss)", (cls._RESOLVED_MANAGER_INTERFACE, "DNS")),
                Gio.DBusCallFlags.NONE,
                -1,
                None
            ).unpack()[0]
        except GLib.Error:
            return

        for (interface, address_family, address) in data:
            if address_family != socket.AF_INET.value:
                continue

            if interface != 0:
                object_path = manager_proxy.call_sync(
                    "GetLink",
                    GLib.Variant("(i)", (interface,)),
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                ).unpack()[0]

                if not bus.call_sync(
                    cls._RESOLVED_NAME,
                    object_path,
                    "org.freedesktop.DBus.Properties",
                    "Get",
                    GLib.Variant("(ss)", ("org.freedesktop.resolve1.Link", "DefaultRoute")),
                    None,
                    Gio.DBusCallFlags.NONE,
                    -1,
                    None
                ).unpack()[0]:
                    continue

            addr = ip_address('.'.join(str(p) for p in address))
            assert isinstance(addr, IPv4Address)
            yield addr

    @classmethod
    def _get_servers_from_resolver(cls) -> Generator[IPv4Address, None, None]:
        with open(cls.RESOLVER_PATH) as f:
            for line in f:
                match = re.search(r"^nameserver\s+((?:\d{1,3}\.){3}\d{1,3}$)", line)
                if match:
                    yield IPv4Address(match.group(1))

    def _subscribe_systemd_resolved(self) -> None:
        def on_signal(
            _connection: Gio.DBusConnection,
            _sender_name: str,
            _object_path: str,
            _interface_name: str,
            _signal_name: str,
            param: GLib.Variant,
        ) -> None:
            interface_name, changed_properties, invalidated_propeties = param.unpack()
            if interface_name == self._RESOLVED_MANAGER_INTERFACE and "DNS" in changed_properties:
                self.emit("changed")

        self._bus = Gio.bus_get_sync(Gio.BusType.SYSTEM)

        self._bus.signal_subscribe(
            self._RESOLVED_NAME, "org.freedesktop.DBus.Properties", "PropertiesChanged",
            "/org/freedesktop/resolve1", None, Gio.DBusSignalFlags.NONE, on_signal)

    def _subscribe_resolver(self) -> None:
        self._monitor = Gio.File.new_for_path(self.RESOLVER_PATH).monitor_file(Gio.FileMonitorFlags.NONE)
        self._monitor.connect("changed", lambda *args: self.emit("changed"))



import weakref

from gi.repository import GObject, GLib
from gi.repository import Gio

from blueman.bluemantyping import GSignals


class AnyBase(GObject.GObject):
    __gsignals__: GSignals = {
        'property-changed': (GObject.SignalFlags.NO_HOOKS, None, (str, object, str))
    }

    connect_signal = GObject.GObject.connect
    disconnect_signal = GObject.GObject.disconnect

    def __init__(self, interface_name: str):
        super().__init__()

        bus = Gio.bus_get_sync(Gio.BusType.SYSTEM)

        this = weakref.proxy(self)

        def on_signal(
            _connection: Gio.DBusConnection,
            _sender_name: str,
            object_path: str,
            _interface_name: str,
            _signal_name: str,
            param: GLib.Variant,
        ) -> None:
            iface_name, changed, invalidated = param.unpack()
            if iface_name == interface_name and this is not None:
                for key in list(changed) + invalidated:
                    this.emit('property-changed', key, changed.get(key, None), object_path)

        weakref.finalize(
            self,
            bus.signal_unsubscribe,
            bus.signal_subscribe(
                "org.bluez",
                "org.freedesktop.DBus.Properties",
                "PropertiesChanged",
                None,
                None,
                Gio.DBusSignalFlags.NONE,
                on_signal
            )
        )
